""" 
This package provides a factory function and utilities for creating ultimate 
bearing capacity calculations using methods like Hansen, Terzaghi, and Vesic 
for various foundation shapes.
"""
import enum
from typing import Optional

from geolysis.foundation import Shape, create_foundation
from geolysis.utils import enum_repr
from geolysis.utils.exceptions import ErrorMsg, ValidationError

from ._core import UltimateBearingCapacity
from .hansen_ubc import HansenUltimateBearingCapacity
from .terzaghi_ubc import (TerzaghiUBC4CircularFooting,
                           TerzaghiUBC4RectangularFooting,
                           TerzaghiUBC4SquareFooting, TerzaghiUBC4StripFooting)
from .vesic_ubc import VesicUltimateBearingCapacity

__all__ = ["UBCType",
           "TerzaghiUBC4StripFooting",
           "TerzaghiUBC4CircularFooting",
           "TerzaghiUBC4RectangularFooting",
           "TerzaghiUBC4SquareFooting",
           "HansenUltimateBearingCapacity",
           "VesicUltimateBearingCapacity",
           "create_ultimate_bearing_capacity"]


@enum_repr
class UBCType(enum.StrEnum):
    """Enumeration of available ultimate bearing capacity types."""
    HANSEN = enum.auto()
    TERZAGHI = enum.auto()
    VESIC = enum.auto()


ubc_classes = {
    UBCType.HANSEN: HansenUltimateBearingCapacity,
    UBCType.TERZAGHI: {Shape.STRIP: TerzaghiUBC4StripFooting,
                       Shape.CIRCLE: TerzaghiUBC4CircularFooting,
                       Shape.SQUARE: TerzaghiUBC4SquareFooting,
                       Shape.RECTANGLE: TerzaghiUBC4RectangularFooting},
    UBCType.VESIC: VesicUltimateBearingCapacity,
}


def create_ultimate_bearing_capacity(friction_angle: float,
                                     cohesion: float,
                                     moist_unit_wgt: float,
                                     depth: float,
                                     width: float,
                                     length: Optional[float] = None,
                                     eccentricity: float = 0.0,
                                     ground_water_level: Optional[
                                         float] = None,
                                     load_angle: float = 0.0,
                                     apply_local_shear: bool = False,
                                     shape: Shape | str = Shape.SQUARE,
                                     ubc_type: Optional[UBCType | str] = None,
                                     ) -> UltimateBearingCapacity:
    r"""A factory function that encapsulate the creation of ultimate bearing
    capacity.

    :param friction_angle: Internal angle of friction for general shear
                           failure (degree).
    :type friction_angle: float

    :param cohesion: Cohesion of soil (:math:`kPa`).
    :type cohesion: float

    :param moist_unit_wgt: Moist unit weight of soil (:math:`kN/m^3`).
    :type moist_unit_wgt: float

    :param depth: Depth of foundation (m).
    :type depth: float

    :param width: Width of foundation footing (m).
    :type width: float

    :param length: Length of foundation footing (m).
    :type length: float, optional

    :param eccentricity: The deviation of the foundation load from the
                         center of gravity of the foundation footing,
                         defaults to 0.0 (m). This means that the foundation
                         load aligns with the center of gravity of the
                         foundation footing.
    :type eccentricity: float, optional

    :param ground_water_level: Depth of water below ground level (m).
    :type ground_water_level: float, optional

    :param load_angle: Inclination of the applied load with the  vertical
                       (:math:`\alpha^{\circ}`), defaults to 0.0.
    :type load_angle: float, optional

    :param apply_local_shear: Indicate whether bearing capacity failure is
                              general or local shear failure, defaults to
                              False.
    :type apply_local_shear: bool, optional

    :param shape: Shape of foundation footing, defaults to
                  :attr:`~geolysis.foundation.Shape.SQUARE`.
    :type shape: Shape | str, optional

    :param ubc_type: Type of allowable bearing capacity calculation to apply.
                     Available values are: :attr:`~UBCType.HANSEN`,
                     :attr:`~UBCType.TERZAGHI`, and :attr:`~UBCType.VESIC`
                     defaults to None.
    :type ubc_type:  UBCType | str, optional

    :raises ValueError: Raised if ubc_type is not supported.
    :raises ValueError: Raised when length is not provided for a rectangular
                        footing.
    :raises ValueError: Raised if an invalid footing shape is provided.
    """

    msg = ErrorMsg(param_name="ubc_type",
                   param_value=ubc_type,
                   symbol="in",
                   param_value_bound=list(UBCType))

    if ubc_type is None:
        raise ValidationError(msg)

    try:
        ubc_type = UBCType(str(ubc_type).casefold())
    except ValueError as e:
        raise ValidationError(msg) from e

    # exception from create_foundation will automatically propagate
    # no need to catch and handle it.
    fnd_size = create_foundation(depth=depth,
                                 width=width,
                                 length=length,
                                 eccentricity=eccentricity,
                                 ground_water_level=ground_water_level,
                                 shape=shape)

    if ubc_type == UBCType.TERZAGHI:
        ubc_class = ubc_classes[ubc_type][fnd_size.footing_shape]
    else:
        ubc_class = ubc_classes[ubc_type]

    ubc = ubc_class(friction_angle=friction_angle,
                    cohesion=cohesion,
                    moist_unit_wgt=moist_unit_wgt,
                    foundation_size=fnd_size,
                    load_angle=load_angle,
                    apply_local_shear=apply_local_shear)
    return ubc



"""validators"""
import operator
from functools import wraps
from typing import Any, Callable, Iterable, TypeAlias

from .exceptions import ErrorMsg, ValidationError

Number: TypeAlias = int | float


class _Validator:

    def __init__(self, bound: Any, /, *,
                 symbol: str,
                 func: Callable,
                 exc_type: Callable,
                 err_msg: str):
        """

        """
        self.bound = bound
        self.symbol = symbol
        self.func = func
        self.exc_type = exc_type
        self.err_msg = err_msg


class _NumValidator(_Validator):

    def _check_val(self, v: Number, fname: str):
        if not self.func(v, self.bound):
            msg = ErrorMsg(msg=self.err_msg,
                           param_name=fname,
                           param_value=v,
                           symbol=self.symbol,
                           param_value_bound=self.bound)
            raise self.exc_type(msg)

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(obj, val):
            if isinstance(val, Iterable):
                for v in val:
                    self._check_val(v, fn.__name__)
            else:
                self._check_val(val, fn.__name__)

            fn(obj, val)

        return wrapper


class _LenValidator(_Validator):
    def _check_val(self, v: Iterable[Any], fname: str):
        if not self.func(len(v), self.bound):
            msg = ErrorMsg(msg=self.err_msg,
                           param_name=fname,
                           param_value=v,
                           symbol=self.symbol,
                           param_value_bound=self.bound)
            msg = "Length of " + msg
            raise self.exc_type(msg)

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(obj, val: Iterable):
            self._check_val(val, fn.__name__)
            fn(obj, val)

        return wrapper


class _InValidator(_Validator):
    def _check_val(self, v, fname):
        if not self.func(self.bound, v):
            msg = ErrorMsg(msg=self.err_msg,
                           param_name=fname,
                           param_value=v,
                           symbol=self.symbol,
                           param_value_bound=self.bound)
            raise self.exc_type(msg)

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(obj, val):
            self._check_val(val, fn.__name__)
            fn(obj, val)

        return wrapper


def in_(val: Iterable[Any], /, *, exc_type=ValidationError, err_msg=None):
    return _InValidator(val, symbol="in", func=operator.contains,
                        exc_type=exc_type, err_msg=err_msg)


def min_len(val: int, /, *, exc_type=ValidationError, err_msg=None):
    return _LenValidator(val, symbol=">=", func=operator.ge,
                         exc_type=exc_type, err_msg=err_msg)


def lt(val: Number, /, *, exc_type=ValidationError, err_msg=None):
    return _NumValidator(val, symbol="<", func=operator.lt,
                         exc_type=exc_type, err_msg=err_msg)


def le(val: Number, /, *, exc_type=ValidationError, err_msg=None):
    return _NumValidator(val, symbol="<=", func=operator.le,
                         exc_type=exc_type, err_msg=err_msg)


def eq(val: Number, /, *, exc_type=ValidationError, err_msg=None):
    return _NumValidator(val, symbol="==", func=operator.eq,
                         exc_type=exc_type, err_msg=err_msg)


def ne(val: Number, /, *, exc_type=ValidationError, err_msg=None):
    return _NumValidator(val, symbol="!=", func=operator.ne,
                         exc_type=exc_type, err_msg=err_msg)


def ge(val: Number, /, *, exc_type=ValidationError, err_msg=None):
    return _NumValidator(val, symbol=">=", func=operator.ge,
                         exc_type=exc_type, err_msg=err_msg)


def gt(val: Number, /, *, exc_type=ValidationError, err_msg=None):
    return _NumValidator(val, symbol=">", func=operator.gt,
                         exc_type=exc_type, err_msg=err_msg)



from abc import ABC

from geolysis.utils import cos, cot, deg2rad, exp, isclose, pi, round_, tan

from ._core import UltimateBearingCapacity

__all__ = ["TerzaghiUBC4StripFooting",
           "TerzaghiUBC4CircularFooting",
           "TerzaghiUBC4SquareFooting",
           "TerzaghiUBC4RectangularFooting"]


@round_(ndigits=2)
def n_c(friction_angle: float) -> float:
    if isclose(friction_angle, 0.0):
        return 5.7
    return cot(friction_angle) * (n_q(friction_angle) - 1.0)


@round_(ndigits=2)
def n_q(friction_angle: float) -> float:
    return (exp((3.0 * pi / 2.0 - deg2rad(friction_angle))
                * tan(friction_angle))
            / (2.0 * (cos(45.0 + friction_angle / 2.0)) ** 2.0))


@round_(ndigits=2)
def n_gamma(friction_angle: float) -> float:
    return (n_q(friction_angle) - 1.0) * tan(1.4 * friction_angle)


class TerzaghiUltimateBearingCapacity(UltimateBearingCapacity, ABC):

    @property
    def n_c(self) -> float:
        r"""Bearing capacity factor :math:`N_c`.

        :Equation:

        .. math:: N_c = \cot(\phi) \cdot (N_q - 1)
        """
        return n_c(self.friction_angle)

    @property
    def n_q(self) -> float:
        r"""Bearing capacity factor :math:`N_q`.

        :Equation:

        .. math::

            N_q = \dfrac{e^{(\frac{3\pi}{2} - \phi)\tan\phi}}
                      {2\cos^2(45 + \frac{\phi}{2})}
        """
        return n_q(self.friction_angle)

    @property
    def n_gamma(self) -> float:
        r"""Bearing capacity factor :math:`N_{\gamma}`.

        :Equation:

        .. math:: N_{\gamma} =  (N_q - 1) \cdot \tan(1.4\phi)
        """
        return n_gamma(self.friction_angle)


class TerzaghiUBC4StripFooting(TerzaghiUltimateBearingCapacity):
    r"""Ultimate bearing capacity for strip footing according to
    ``Terzaghi 1943``.

    :Equation:

    .. math:: q_u = cN_c + qN_q + 0.5 \gamma BN_{\gamma}

    .. list-table::
       :widths: auto
       :header-rows: 1

       * - Symbol
         - Description
         - Unit
       * - :math:`q_u`
         - Ultimate bearing capacity
         - :math:`kPa`
       * - :math:`c`
         - Cohesion of soil
         - :math:`kPa`
       * - :math:`q`
         - Overburden pressure of soil
         - :math:`kPa`
       * - :math:`\gamma`
         - Unit weight of soil
         - :math:`kN/m^3`
       * - :math:`B`
         - Width of foundation footing
         - :math:`m`
       * - :math:`N_c`, :math:`N_q`, :math:`N_{\gamma}`
         - Bearing capacity factors
         - —
    """

    @round_(ndigits=2)
    def bearing_capacity(self) -> float:
        """Calculates ultimate bearing capacity for strip footing."""
        return (self._cohesion_term(1.0)
                + self._surcharge_term()
                + self._embedment_term(0.5))


class TerzaghiUBC4CircularFooting(TerzaghiUltimateBearingCapacity):
    r"""Ultimate bearing capacity for circular footing according to
    ``Terzaghi 1943``.

    :Equation:

    .. math:: q_u = 1.3cN_c + qN_q + 0.3 \gamma BN_{\gamma}

    .. list-table::
       :widths: auto
       :header-rows: 1

       * - Symbol
         - Description
         - Unit
       * - :math:`q_u`
         - Ultimate bearing capacity
         - :math:`kPa`
       * - :math:`c`
         - Cohesion of soil
         - :math:`kPa`
       * - :math:`q`
         - Overburden pressure of soil
         - :math:`kPa`
       * - :math:`\gamma`
         - Unit weight of soil
         - :math:`kN/m^3`
       * - :math:`B`
         - Width of foundation footing
         - :math:`m`
       * - :math:`N_c`, :math:`N_q`, :math:`N_{\gamma}`
         - Bearing capacity factors
         - —
    """

    @round_(ndigits=2)
    def bearing_capacity(self) -> float:
        """Calculates ultimate bearing capacity for circular footing."""
        return (self._cohesion_term(1.3)
                + self._surcharge_term()
                + self._embedment_term(0.3))


class TerzaghiUBC4RectangularFooting(TerzaghiUltimateBearingCapacity):
    r"""Ultimate bearing capacity for rectangular footing according to
    ``Terzaghi 1943``.

    :Equation:

    .. math::

            q_u = \left(1 + 0.3 \dfrac{B}{L} \right) c N_c + qN_q
                  + \left(1 - 0.2 \dfrac{B}{L} \right) 0.5 B \gamma N_{\gamma}

    .. list-table::
       :widths: auto
       :header-rows: 1

       * - Symbol
         - Description
         - Unit
       * - :math:`q_u`
         - Ultimate bearing capacity
         - :math:`kPa`
       * - :math:`c`
         - Cohesion of soil
         - :math:`kPa`
       * - :math:`q`
         - Overburden pressure of soil
         - :math:`kPa`
       * - :math:`\gamma`
         - Unit weight of soil
         - :math:`kN/m^3`
       * - :math:`B`
         - Width of foundation footing
         - :math:`m`
       * - :math:`L`
         - Length of foundation footing
         - :math:`m`
       * - :math:`N_c`, :math:`N_q`, :math:`N_{\gamma}`
         - Bearing capacity factors
         - —
    """

    @round_(ndigits=2)
    def bearing_capacity(self) -> float:
        """Calculates ultimate bearing capacity for rectangular footing."""
        width = self.foundation_size.width
        length = self.foundation_size.length
        coh_coef = 1.0 + 0.3 * (width / length)
        emb_coef = (1.0 - 0.2 * (width / length)) / 2.0

        return (self._cohesion_term(coh_coef)
                + self._surcharge_term()
                + self._embedment_term(emb_coef))


class TerzaghiUBC4SquareFooting(TerzaghiUBC4RectangularFooting):
    r"""Ultimate bearing capacity for square footing according to 
    ``Terzaghi 1943``.

    :Equation:

    .. math:: q_u = 1.3cN_c + qN_q + 0.4 \gamma BN_{\gamma}

    .. list-table::
       :widths: auto
       :header-rows: 1

       * - Symbol
         - Description
         - Unit
       * - :math:`q_u`
         - Ultimate bearing capacity
         - :math:`kPa`
       * - :math:`c`
         - Cohesion of soil
         - :math:`kPa`
       * - :math:`q`
         - Overburden pressure of soil
         - :math:`kPa`
       * - :math:`\gamma`
         - Unit weight of soil
         - :math:`kN/m^3`
       * - :math:`B`
         - Width of foundation footing
         - :math:`m`
       * - :math:`N_c`, :math:`N_q`, :math:`N_{\gamma}`
         - Bearing capacity factors
         - —
    """

    def bearing_capacity(self):
        """Calcalates ultimate bearing capacity for square footing.
        """
        return super().bearing_capacity()



from geolysis.foundation import Shape
from geolysis.utils import isclose, round_, sin, tan

from . import hansen_ubc
from ._core import UltimateBearingCapacity

__all__ = ["VesicUltimateBearingCapacity"]


@round_(ndigits=2)
def n_c(friction_angle: float) -> float:
    return hansen_ubc.n_c(friction_angle)


@round_(ndigits=2)
def n_q(friction_angle: float) -> float:
    return hansen_ubc.n_q(friction_angle)


@round_(ndigits=2)
def n_gamma(friction_angle: float) -> float:
    return 2.0 * (n_q(friction_angle) + 1.0) * tan(friction_angle)


@round_(ndigits=2)
def s_c(friction_angle: float,
        f_width: float,
        f_length: float,
        f_shape: Shape) -> float:
    _n_q = n_q(friction_angle)
    _n_c = n_c(friction_angle)

    if f_shape == Shape.STRIP:
        return 1.0
    elif f_shape == Shape.RECTANGLE:
        return 1.0 + (f_width / f_length) * (_n_q / _n_c)
    else:  # SQUARE, CIRCLE
        return 1.0 + (_n_q / _n_c)


@round_(ndigits=2)
def s_q(friction_angle: float,
        f_width: float,
        f_length: float,
        f_shape: Shape) -> float:
    if f_shape == Shape.STRIP:
        return 1.0
    elif f_shape == Shape.RECTANGLE:
        return 1.0 + (f_width / f_length) * tan(friction_angle)
    else:  # SQUARE, CIRCLE
        return 1.0 + tan(friction_angle)


@round_(ndigits=2)
def s_gamma(f_width: float, f_length: float, f_shape: Shape) -> float:
    if f_shape == Shape.STRIP:
        return 1.0
    elif f_shape == Shape.RECTANGLE:
        return 1.0 - 0.4 * (f_width / f_length)
    else:  # SQUARE, CIRCLE
        return 0.6


@round_(ndigits=2)
def d_c(f_depth: float, f_width: float) -> float:
    return 1.0 + 0.4 * f_depth / f_width


@round_(ndigits=2)
def d_q(friction_angle: float, f_depth: float, f_width: float) -> float:
    return (1.0 + 2.0 * tan(friction_angle)
            * (1.0 - sin(friction_angle)) ** 2.0
            * (f_depth / f_width))


@round_(ndigits=2)
def d_gamma() -> float:
    return 1.0


@round_(ndigits=2)
def i_c(load_angle: float) -> float:
    return (1.0 - load_angle / 90.0) ** 2.0


@round_(ndigits=2)
def i_q(load_angle: float) -> float:
    return i_c(load_angle)


@round_(ndigits=2)
def i_gamma(friction_angle: float, load_angle: float) -> float:
    if isclose(friction_angle, 0.0):
        return 1.0
    return (1.0 - load_angle / friction_angle) ** 2.0


class VesicUltimateBearingCapacity(UltimateBearingCapacity):
    r"""Ultimate bearing capacity for soils according to ``Vesic (1973)``.

    :Equation:

    .. math::

            q_u = cN_c s_c d_c i_c + qN_q s_q d_q i_q
                  + 0.5 \gamma B N_{\gamma} s_{\gamma} d_{\gamma} i_{\gamma}

    .. list-table::
       :widths: auto
       :header-rows: 1

       * - Symbol
         - Description
         - Unit
       * - :math:`q_u`
         - Ultimate bearing capacity
         - :math:`kPa`
       * - :math:`c`
         - Cohesion of soil
         - :math:`kPa`
       * - :math:`q`
         - Overburden pressure of soil
         - :math:`kPa`
       * - :math:`\gamma`
         - Unit weight of soil
         - :math:`kN/m^3`
       * - :math:`B`
         - Width of foundation footing
         - :math:`m`
       * - :math:`N_c`, :math:`N_q`, :math:`N_{\gamma}`
         - Bearing capacity factors
         - —
       * - :math:`s_c`, :math:`s_q`, :math:`s_{\gamma}`
         - Shape factors
         - —
       * - :math:`d_c`, :math:`d_q`, :math:`d_{\gamma}`
         - Depth factors
         - —
       * - :math:`i_c`, :math:`i_q`, :math:`i_{\gamma}`
         - Inclination factors
         - —
    """

    @property
    def n_c(self) -> float:
        r"""Bearing capacity factor :math:`N_c`.

        :Equation:

        .. math:: N_c = \cot(\phi) \left(N_q - 1\right)
        """
        return n_c(self.friction_angle)

    @property
    def n_q(self) -> float:
        r"""Bearing capacity factor :math:`N_q`.

        :Equation:

        .. math:: N_q = \tan^2\left(45 + \frac{\phi}{2}\right) \cdot
                      e^{\pi \tan(\phi)}
        """
        return n_q(self.friction_angle)

    @property
    def n_gamma(self) -> float:
        r"""Bearing capacity factor :math:`N_{\gamma}`.

        :Equation:

        .. math:: N_{\gamma} = 2(N_q + 1) \tan(\phi)
        """
        return n_gamma(self.friction_angle)

    @property
    def s_c(self) -> float:
        r"""Shape factor :math:`S_c`.

        :Equation:

        .. math::

             s_c &= 1.0 \rightarrow \text{Strip footing}

             s_c &= 1 + \dfrac{B}{L} \cdot \dfrac{N_q}{N_c} \rightarrow
                     \text{Rectangular footing}

             s_c &= 1 + \dfrac{N_q}{N_c} \rightarrow
                       \text{Square or circular footing}
        """
        width, length, shape = self.foundation_size.footing_params()
        return s_c(self.friction_angle, width, length, shape)

    @property
    def s_q(self) -> float:
        r"""Shape factor :math:`S_q`.

        :Equation:

        .. math::

             s_q &= 1.0 \rightarrow \text{Strip footing}

             s_q &= 1 + \dfrac{B}{L} \cdot \tan(\phi) \rightarrow
                    \text{Rectangular footing}

             s_q &= 1 + \tan(\phi) \rightarrow \text{Square or circular footing}
        """
        width, length, shape = self.foundation_size.footing_params()
        return s_q(self.friction_angle, width, length, shape)

    @property
    def s_gamma(self) -> float:
        r"""Shape factor :math:`S_{\gamma}`.

        :Equation:

        .. math::

             s_{\gamma} &= 1.0 \rightarrow \text{Strip footing}

             s_{\gamma} &= 1.0 - 0.4 \dfrac{B}{L} \rightarrow
                          \text{Rectangular footing}

             s_{\gamma} &= 0.6 \rightarrow \text{Square or circular footing}
        """
        width, length, shape = self.foundation_size.footing_params()
        return s_gamma(width, length, shape)

    @property
    def d_c(self) -> float:
        r"""Depth factor :math:`D_c`.

        :Equation:

        .. math:: d_c = 1 + 0.4 \dfrac{D_f}{B}
        """
        depth, width = self.foundation_size.depth, self.foundation_size.width
        return d_c(depth, width)

    @property
    def d_q(self) -> float:
        r"""Depth factor :math:`D_q`.

        :Equation:

        .. math::

            d_q = 1 + 2 \tan(\phi) \cdot (1 - \sin(\phi))^2
                  \cdot \dfrac{D_f}{B}
        """
        depth, width = self.foundation_size.depth, self.foundation_size.width
        return d_q(self.friction_angle, depth, width)

    @property
    def d_gamma(self) -> float:
        r"""Depth factor :math:`D_{\gamma}`.

        :Equation:

        .. math:: d_{\gamma} = 1.0
        """
        return d_gamma()

    @property
    def i_c(self) -> float:
        r"""Inclination factor :math:`I_c`.

        :Equation:

        .. math:: i_c = (1 - \dfrac{\alpha}{90})^2
        """
        return i_c(self.load_angle)

    @property
    def i_q(self) -> float:
        r"""Inclination factor :math:`I_q`.

        :Equation:

        .. math:: i_q = (1 - \dfrac{\alpha}{90})^2
        """
        return i_q(self.load_angle)

    @property
    def i_gamma(self) -> float:
        r"""Inclination factor :math:`I_{\gamma}`.

        :Equation:

        .. math:: i_{\gamma} = \left(1 - \dfrac{\alpha}{\phi} \right)^2
        """
        return i_gamma(self.friction_angle, self.load_angle)



from geolysis.foundation import Shape
from geolysis.utils import cos, cot, exp, isclose, pi, round_, sin, tan

from ._core import UltimateBearingCapacity

__all__ = ["HansenUltimateBearingCapacity"]


@round_(ndigits=2)
def n_c(friction_angle: float) -> float:
    if isclose(friction_angle, 0.0):
        return 5.14
    return cot(friction_angle) * (n_q(friction_angle) - 1.0)


@round_(ndigits=2)
def n_q(friction_angle: float) -> float:
    return (tan(45.0 + friction_angle / 2.0) ** 2.0
            * exp(pi * tan(friction_angle)))


@round_(ndigits=2)
def n_gamma(friction_angle: float) -> float:
    return 1.8 * (n_q(friction_angle) - 1.0) * tan(friction_angle)


@round_(ndigits=2)
def s_c(f_width: float, f_length: float, f_shape: Shape) -> float:
    if f_shape == Shape.STRIP:
        return 1.0
    elif f_shape == Shape.RECTANGLE:
        return 1.0 + 0.2 * f_width / f_length
    else:  # SQUARE & CIRCLE
        return 1.3


@round_(ndigits=2)
def s_q(f_width: float, f_length: float, f_shape: Shape) -> float:
    if f_shape == Shape.STRIP:
        return 1.0
    elif f_shape == Shape.RECTANGLE:
        return 1.0 + 0.2 * f_width / f_length
    else:  # SQUARE & CIRCLE
        return 1.2


@round_(ndigits=2)
def s_gamma(f_width: float, f_length: float, f_shape: Shape) -> float:
    if f_shape == Shape.STRIP:
        return 1.0
    elif f_shape == Shape.RECTANGLE:
        return 1.0 - 0.4 * f_width / f_length
    elif f_shape == Shape.SQUARE:
        return 0.8
    else:  # CIRCLE
        return 0.6


@round_(ndigits=2)
def d_c(f_depth: float, f_width: float) -> float:
    return 1.0 + 0.35 * f_depth / f_width


@round_(ndigits=2)
def d_q(f_depth: float, f_width: float) -> float:
    return d_c(f_depth, f_width)


@round_(ndigits=2)
def d_gamma() -> float:
    return 1.0


@round_(ndigits=2)
def i_c(cohesion: float,
        load_angle: float,
        f_width: float,
        f_length: float) -> float:
    return 1.0 - sin(load_angle) / (2.0 * cohesion * f_width * f_length)


@round_(ndigits=2)
def i_q(load_angle: float) -> float:
    return 1.0 - (1.5 * sin(load_angle)) / cos(load_angle)


@round_(ndigits=2)
def i_gamma(load_angle: float) -> float:
    return i_q(load_angle) ** 2.0


class HansenUltimateBearingCapacity(UltimateBearingCapacity):
    r"""Ultimate bearing capacity for soils according to ``Hansen (1961)``.

    :Equation:

    .. math::

            q_u = cN_c s_c d_c i_c + qN_q s_q d_q i_q
                  + 0.5 \gamma B N_{\gamma} s_{\gamma} d_{\gamma} i_{\gamma}

    .. list-table::
       :widths: auto
       :header-rows: 1

       * - Symbol
         - Description
         - Unit
       * - :math:`q_u`
         - Ultimate bearing capacity
         - :math:`kPa`
       * - :math:`c`
         - Cohesion of soil
         - :math:`kPa`
       * - :math:`q`
         - Overburden pressure of soil
         - :math:`kPa`
       * - :math:`\gamma`
         - Unit weight of soil
         - :math:`kN/m^3`
       * - :math:`B`
         - Width of foundation footing
         - :math:`m`
       * - :math:`N_c`, :math:`N_q`, :math:`N_{\gamma}`
         - Bearing capacity factors
         - —
       * - :math:`s_c`, :math:`s_q`, :math:`s_{\gamma}`
         - Shape factors
         - —
       * - :math:`d_c`, :math:`d_q`, :math:`d_{\gamma}`
         - Depth factors
         - —
       * - :math:`i_c`, :math:`i_q`, :math:`i_{\gamma}`
         - Inclination factors
         - —
    """

    @property
    def n_c(self) -> float:
        r"""Bearing capacity factor :math:`N_c`.

        :Equation:

        .. math:: N_c = \cot(\phi) \left(N_q - 1\right)
        """
        return n_c(self.friction_angle)

    @property
    def n_q(self) -> float:
        r"""Bearing capacity factor :math:`N_q`.

        :Equation:

        .. math::

            N_q = \tan^2\left(45 + \frac{\phi}{2}\right) \cdot e^{\pi \tan(\phi)}
        """
        return n_q(self.friction_angle)

    @property
    def n_gamma(self) -> float:
        r"""Bearing capacity factor :math:`N_{\gamma}`.

        :Equation:

        .. math:: N_{\gamma} = 1.8 \left(N_q - 1\right) \tan(\phi)
        """
        return n_gamma(self.friction_angle)

    @property
    def s_c(self) -> float:
        r"""Shape factor :math:`S_c`.

        :Equation:

        .. math::

            s_c &= 1.0 \rightarrow \text{Strip footing}

            s_c &= 1.0 + 0.2 \frac{B}{L} \rightarrow \text{Rectangular footing}

            s_c &= 1.3 \rightarrow \text{Square or circular footing}
        """
        width, length, shape = self.foundation_size.footing_params()
        return s_c(width, length, shape)

    @property
    def s_q(self) -> float:
        r"""Shape factor :math:`S_q`.

        :Equation:

        .. math::

            s_q &= 1.0 \rightarrow \text{Strip footing}

            s_q &= 1.0 + 0.2 \frac{B}{L} \rightarrow \text{Rectangular footing}

            s_q &= 1.2 \rightarrow \text{Square or circular footing}
        """
        width, length, shape = self.foundation_size.footing_params()
        return s_q(width, length, shape)

    @property
    def s_gamma(self) -> float:
        r"""Shape factor :math:`S_{\gamma}`.

        :Equation:

        .. math::

            s_{\gamma} &= 1.0 \rightarrow \text{Strip footing}

            s_{\gamma} &= 1.0 - 0.4 \frac{B}{L} \rightarrow
                          \text{Rectangular footing}

            s_{\gamma} &= 0.8 \rightarrow \text{Square footing}

            s_{\gamma} &= 0.6 \rightarrow \text{Circular footing}
        """
        width, length, shape = self.foundation_size.footing_params()
        return s_gamma(width, length, shape)

    @property
    def d_c(self) -> float:
        r"""Depth factor :math:`D_c`.

        :Equation:

        .. math:: d_c = 1.0 + 0.35 \cdot \frac{D_f}{B}
        """
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        return d_c(depth, width)

    @property
    def d_q(self) -> float:
        r"""Depth factor :math:`D_q`.

        :Equation:

        .. math:: d_q = 1.0 + 0.35 \cdot \frac{D_f}{B}
        """
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        return d_q(depth, width)

    @property
    def d_gamma(self) -> float:
        r"""Depth factor :math:`D_{\gamma}`.

        :Equation:

        .. math:: d_{\gamma} = 1.0
        """
        return d_gamma()

    @property
    def i_c(self) -> float:
        r"""Inclination factor :math:`I_c`.

        :Equation:

        .. math:: I_c = 1 - \frac{\sin(\alpha)}{2cBL}
        """
        width, length = self.foundation_size.width, self.foundation_size.length
        return i_c(self.cohesion, self.load_angle, width, length)

    @property
    def i_q(self) -> float:
        r"""Inclination factor :math:`I_q`.

        :Equation:

        .. math:: I_q = 1 - \frac{1.5 \sin(\alpha)}{\cos(\alpha)}
        """
        return i_q(self.load_angle)

    @property
    def i_gamma(self) -> float:
        r"""Inclination factor :math:`I_{\gamma}`.

        :Equation:

        .. math:: I_{\gamma} = I_q^2
        """
        return i_gamma(self.load_angle)



from abc import ABC, abstractmethod

from geolysis.foundation import FoundationSize
from geolysis.utils import arctan, round_, tan, validators


class UltimateBearingCapacity(ABC):
    def __init__(self, friction_angle: float,
                 cohesion: float,
                 moist_unit_wgt: float,
                 foundation_size: FoundationSize,
                 load_angle: float = 0.0,
                 apply_local_shear: bool = False) -> None:
        r"""
        :param friction_angle: Internal angle of friction for general shear 
                               failure (degrees).
        :type friction_angle: float

        :param cohesion: Cohesion of soil (:math:`kPa`).
        :type cohesion: float

        :param moist_unit_wgt: Moist unit weight of soil (:math:`kN/m^3`).
        :type moist_unit_wgt: float

        :param foundation_size: Size of the foundation.
        :type foundation_size: FoundationSize

        :param load_angle: Inclination of the applied load with the  vertical 
                           (:math:`\alpha^{\circ}`), defaults to 0.0.
        :type load_angle: float, optional

        :param apply_local_shear: Indicate whether bearing capacity failure is
                                  general shear or local shear failure,
                                  defaults to False.
        :type apply_local_shear: bool, optional
        """
        self.friction_angle = friction_angle
        self.cohesion = cohesion
        self.moist_unit_wgt = moist_unit_wgt
        self.load_angle = load_angle
        self.foundation_size = foundation_size
        self.apply_local_shear = apply_local_shear

    @property
    def friction_angle(self) -> float:
        r"""Return friction angle for local shear in the case of local shear
        failure or general shear in the case of general shear failure.

        :Equation:

        In the case of local shear failure:

        .. math::

           \phi' = \tan^{-1} \left(\frac{2}{3} \tan \phi\right)

        """
        if self.apply_local_shear:
            return arctan((2.0 / 3.0) * tan(self._friction_angle))
        return self._friction_angle

    @friction_angle.setter
    @validators.ge(0.0)
    def friction_angle(self, val: float):
        self._friction_angle = val

    @property
    def cohesion(self) -> float:
        r"""Return cohesion for local shear in the case of local shear failure
        or general shear in the case of general shear failure.

        :Equation:

        In the case of local shear failure:

        .. math::

            C^{'} = \dfrac{2}{3} \cdot C
        """
        if self.apply_local_shear:
            return (2.0 / 3.0) * self._cohesion
        return self._cohesion

    @cohesion.setter
    @validators.ge(0.0)
    def cohesion(self, val: float):
        self._cohesion = val

    @property
    def moist_unit_wgt(self) -> float:
        """Moist unit weight of soil (:math:`kN/m^3`)."""
        return self._moist_unit_wgt

    @moist_unit_wgt.setter
    @validators.gt(0.0)
    def moist_unit_wgt(self, val: float):
        self._moist_unit_wgt = val

    @property
    def load_angle(self) -> float:
        """Inclination of the applied load with the  vertical."""
        return self._load_angle

    @load_angle.setter
    @validators.le(90.0)
    @validators.ge(0.0)
    def load_angle(self, val: float):
        self._load_angle = val

    @property
    def s_c(self) -> float:
        """Shape factor :math:`S_c`"""
        return 1.0

    @property
    def s_q(self) -> float:
        """Shape factor :math:`S_q`"""
        return 1.0

    @property
    def s_gamma(self) -> float:
        r"""Shape factor :math:`S_{\gamma}`"""
        return 1.0

    @property
    def d_c(self) -> float:
        """Depth factor :math:`d_c`"""
        return 1.0

    @property
    def d_q(self) -> float:
        """Depth factor :math:`d_q`"""
        return 1.0

    @property
    def d_gamma(self) -> float:
        r"""Depth factor :math:`d_{\gamma}`"""
        return 1.0

    @property
    def i_c(self) -> float:
        """Inclination factor :math:`i_c`"""
        return 1.0

    @property
    def i_q(self) -> float:
        """Inclination factor :math:`i_q`"""
        return 1.0

    @property
    def i_gamma(self) -> float:
        r"""Inclination factor :math:`i_{\gamma}`"""
        return 1.0

    def _cohesion_term(self, coef: float = 1.0) -> float:
        return coef * self.cohesion * self.n_c * self.s_c * self.d_c * self.i_c

    def _surcharge_term(self) -> float:
        depth = self.foundation_size.depth
        water_level = self.foundation_size.ground_water_level

        if water_level is None:
            water_corr = 1.0  # water correction
        else:
            # water level above the base of the foundation
            a = max(depth - water_level, 0.0)
            water_corr = min(1 - 0.5 * a / depth, 1)

        # effective overburden pressure (surcharge)
        eop = self.moist_unit_wgt * depth
        return eop * self.n_q * self.s_q * self.d_q * self.i_q * water_corr

    def _embedment_term(self, coef: float = 0.5) -> float:
        depth = self.foundation_size.depth
        width = self.foundation_size.effective_width
        water_level = self.foundation_size.ground_water_level

        if water_level is None:
            # water correction
            water_corr = 1.0
        else:
            #: b -> water level below the base of the foundation
            b = max(water_level - depth, 0)
            water_corr = min(0.5 + 0.5 * b / width, 1)

        return (coef * self.moist_unit_wgt * width * self.n_gamma
                * self.s_gamma * self.d_gamma * self.i_gamma * water_corr)

    @round_(ndigits=2)
    def bearing_capacity(self) -> float:
        """Calculates the ultimate bearing capacity."""
        return (self._cohesion_term(1.0)
                + self._surcharge_term()
                + self._embedment_term(0.5))

    @property
    @abstractmethod
    def n_c(self) -> float:
        ...

    @property
    @abstractmethod
    def n_q(self) -> float:
        ...

    @property
    @abstractmethod
    def n_gamma(self) -> float:
        ...



"""This module provides classes for performing Standard Penetration Test (SPT) 
corrections, including energy, overburden pressure, and dilatancy corrections, 
as well as calculating design N-values.
"""
import enum
from abc import abstractmethod
from typing import Final, Literal, Sequence

from .utils import enum_repr, isclose, log10, mean, round_, sqrt, validators
from .utils.exceptions import ErrorMsg, ValidationError

__all__ = ["SPTNDesign",
           "HammerType",
           "SamplerType",
           "EnergyCorrection",
           "GibbsHoltzOPC",
           "BazaraaPeckOPC",
           "PeckOPC",
           "LiaoWhitmanOPC",
           "SkemptonOPC",
           "DilatancyCorrection",
           "OPCType",
           "create_overburden_pressure_correction"]


class SPTNDesign:
    """SPT Design Calculations.

    Due to uncertainty in field procedure in standard penetration test and also
    to consider all the N-value in the influence zone of a foundation, a method
    was suggested to calculate the design N-value which should be used in
    calculating the allowable bearing capacity of shallow foundation rather
    than using a particular N-value. All the N-value from the influence zone is
    taken under consideration by giving the highest weightage to the closest
    N-value from the base.
    """

    def __init__(self, corrected_spt_n_values: Sequence[float],
                 method: Literal["min", "avg", "wgt"] = "wgt") -> None:
        """
        :param corrected_spt_n_values: Corrected SPT N-values within the
                                       foundation influence zone.
        :type corrected_spt_n_values: Sequence[float]
        """
        self.corrected_spt_n_values = corrected_spt_n_values
        self.method = method

    @property
    def corrected_spt_n_values(self) -> Sequence[float]:
        """Corrected SPT N-values within the foundation influence zone."""
        return self._corrected_spt_n_values

    @corrected_spt_n_values.setter
    @validators.le(100.0)
    @validators.gt(0.0)
    @validators.min_len(1)
    def corrected_spt_n_values(self, val: Sequence[float]) -> None:
        self._corrected_spt_n_values = val

    @property
    def method(self):
        return self._method

    @method.setter
    @validators.in_(("min", "avg", "wgt"))
    def method(self, val: str) -> None:
        self._method = val

    @staticmethod
    def _avg_spt_n_design(vals) -> float:
        return mean(vals)

    @staticmethod
    def _min_spt_n_design(vals):
        return min(vals)

    @staticmethod
    def _wgt_spt_n_design(vals):

        sum_total = 0.0
        sum_wgts = 0.0

        for i, corr_spt_n_val in enumerate(vals, start=1):
            wgt = 1 / i ** 2
            sum_total += wgt * corr_spt_n_val
            sum_wgts += wgt

        return sum_total / sum_wgts

    @round_(ndigits=1)
    def n_design(self):
        r"""Calculates the SPT N-design within the foundation influence zone.

        If ``method="min"``, it returns the minimum N-value within the foundation
        influence zone as the SPT N-design value. This approach was suggested
        by ``Terzaghi & Peck (1948)``.

        if ``method="avg"``, it returns the average N-value within the foundation
        influence zone as the SPT N-design value.

        if ``method="wgt"``, it returns the weighted average N-value within the
        foundation influence zone as the SPT N-design value.

        :Equation:

        .. math::

            N_{design} = \dfrac{\sum_{i=1}^{n} \frac{N_i}{i^2}}
                         {\sum_{i=1}^{n}\frac{1}{i^2}}
        """
        if self.method == "min":
            return self._min_spt_n_design(self.corrected_spt_n_values)
        elif self.method == "avg":
            return self._avg_spt_n_design(self.corrected_spt_n_values)
        else:  # method="wgt"
            return self._wgt_spt_n_design(self.corrected_spt_n_values)


@enum_repr
class HammerType(enum.StrEnum):
    """Enumeration of hammer types."""
    AUTOMATIC = enum.auto()
    DONUT_1 = enum.auto()
    DONUT_2 = enum.auto()
    SAFETY = enum.auto()
    DROP = enum.auto()
    PIN = enum.auto()


@enum_repr
class SamplerType(enum.StrEnum):
    """Enumeration of sampler types."""
    STANDARD = enum.auto()
    NON_STANDARD = enum.auto()


class EnergyCorrection:
    r"""SPT N-value standardized for field procedures.

    On the basis of field observations, it appears reasonable to standardize
    the field SPT N-value as a function of the input driving energy and its
    dissipation around the sampler around the surrounding soil. The variations
    in testing procedures may be at least partially compensated by converting
    the measured N-value to :math:`N_{60}` assuming 60% hammer energy being
    transferred to the tip of the standard split spoon.
    """

    _HAMMER_EFFICIENCY_FACTORS = {HammerType.AUTOMATIC: 0.70,
                                  HammerType.DONUT_1: 0.60,
                                  HammerType.DONUT_2: 0.50,
                                  HammerType.SAFETY: 0.55,
                                  HammerType.DROP: 0.45,
                                  HammerType.PIN: 0.45}

    _SAMPLER_CORRECTION_FACTORS = {SamplerType.STANDARD: 1.00,
                                   SamplerType.NON_STANDARD: 1.20}

    def __init__(self, recorded_spt_n_value: int, *,
                 energy_percentage=0.6,
                 borehole_diameter=65.0,
                 rod_length=3.0,
                 hammer_type=HammerType.DONUT_1,
                 sampler_type=SamplerType.STANDARD):
        """
        :param recorded_spt_n_value: Recorded SPT N-value from field.
        :type recorded_spt_n_value: int

        :param energy_percentage: Energy percentage reaching the tip of the 
                                  sampler, defaults to 0.6
        :type energy_percentage: float, optional

        :param borehole_diameter: Borehole diameter (mm), defaults to 65.0.
        :type borehole_diameter: float, optional

        :param rod_length: Length of SPT rod, defaults to 3.0. (m)
        :type rod_length: float, optional

        :param hammer_type: Hammer type, defaults to :py:enum:mem:`~HammerType.DONUT_1`
        :type hammer_type: HammerType, optional

        :param sampler_type: Sampler type, defaults to :py:enum:mem:`~SamplerType.STANDARD`
        :type sampler_type: SamplerType, optional
        """
        self.recorded_spt_n_value = int(recorded_spt_n_value)
        self.energy_percentage = energy_percentage
        self.borehole_diameter = borehole_diameter
        self.rod_length = rod_length
        self.hammer_type = hammer_type
        self.sampler_type = sampler_type

    @property
    def recorded_spt_n_value(self) -> int:
        """Recorded SPT N-value from field."""
        return self._recorded_spt_value

    @recorded_spt_n_value.setter
    @validators.le(100)
    @validators.gt(0)
    def recorded_spt_n_value(self, val: int) -> None:
        self._recorded_spt_value = val

    @property
    def energy_percentage(self) -> float:
        """Energy percentage reaching the tip of the sampler."""
        return self._energy_percentage

    @energy_percentage.setter
    @validators.le(1.0)
    @validators.gt(0.0)
    def energy_percentage(self, val: float) -> None:
        self._energy_percentage = val

    @property
    def borehole_diameter(self) -> float:
        """Borehole diameter (mm)."""
        return self._borehole_diameter

    @borehole_diameter.setter
    @validators.le(200.0)
    @validators.ge(65.0)
    def borehole_diameter(self, val: float) -> None:
        self._borehole_diameter = val

    @property
    def rod_length(self) -> float:
        """Length of SPT rod."""
        return self._rod_length

    @rod_length.setter
    @validators.gt(0.0)
    def rod_length(self, val: float) -> None:
        self._rod_length = val

    @property
    def hammer_type(self) -> HammerType:
        return self._hammer_type

    @hammer_type.setter
    @validators.in_(tuple(HammerType))
    def hammer_type(self, val: HammerType) -> None:
        self._hammer_type = val

    @property
    def sampler_type(self) -> SamplerType:
        return self._sampler_type

    @sampler_type.setter
    @validators.in_(tuple(SamplerType))
    def sampler_type(self, val: SamplerType) -> None:
        self._sampler_type = val

    @property
    def hammer_efficiency(self) -> float:
        """Hammer efficiency correction factor."""
        return self._HAMMER_EFFICIENCY_FACTORS[self.hammer_type]

    @property
    def borehole_diameter_correction(self) -> float:
        """Borehole diameter correction factor."""
        if 65 <= self.borehole_diameter <= 115:
            corr = 1.00
        elif 115 < self.borehole_diameter <= 150:
            corr = 1.05
        else:
            corr = 1.15
        return corr

    @property
    def sampler_correction(self) -> float:
        """Sampler correction factor."""
        return self._SAMPLER_CORRECTION_FACTORS[self.sampler_type]

    @property
    def rod_length_correction(self) -> float:
        """Rod length correction factor."""
        if 3.0 <= self.rod_length <= 4.0:
            corr = 0.75
        elif 4.0 < self.rod_length <= 6.0:
            corr = 0.85
        elif 6.0 < self.rod_length <= 10.0:
            corr = 0.95
        else:
            corr = 1.00
        return corr

    def correction(self) -> float:
        r"""Energy correction factor.

        :Equation:

        .. math::

            N_{ENERGY} = \dfrac{E_H \cdot C_B \cdot C_S \cdot C_R \cdot N}{ENERGY}

        ``ENERGY``: 0.6, 0.55, etc
        """
        numerator = (self.hammer_efficiency
                     * self.borehole_diameter_correction
                     * self.sampler_correction
                     * self.rod_length_correction)
        return numerator / self.energy_percentage

    @round_(ndigits=1)
    def standardized_spt_n_value(self) -> float:
        """Standardized SPT N-value."""
        return self.correction() * self.recorded_spt_n_value


class OPC:
    """Base class for Overburden Pressure Correction (OPC)."""

    def __init__(self, std_spt_n_value: float, eop: float) -> None:
        """
        :param std_spt_n_value: SPT N-value standardized for field procedures.
        :type std_spt_n_value: float

        :param eop: Effective overburden pressure (:math:`kPa`).
        :type eop: float
        """
        self.std_spt_n_value = std_spt_n_value
        self.eop = eop

    @property
    def std_spt_n_value(self) -> float:
        """SPT N-value standardized for field procedures."""
        return self._std_spt_n_value

    @std_spt_n_value.setter
    @validators.le(100)
    @validators.gt(0.0)
    def std_spt_n_value(self, val: float) -> None:
        self._std_spt_n_value = val

    @round_(ndigits=1)
    def corrected_spt_n_value(self) -> float:
        r"""Corrected SPT N-value.

        :Equation:

        .. math:: (N_1)_{60} = C_N \cdot N_{60}

        .. note::

            ``60`` is used in this case to represent ``60%`` hammer efficiency
            and can be any percentage of hammer efficiency e.g :math:`N_{55}`
            for ``55%`` hammer efficiency.
        """
        corrected_spt = self.correction() * self.std_spt_n_value
        # Corrected SPT should not be more
        # than 2 times the Standardized SPT
        return min(corrected_spt, 2 * self.std_spt_n_value)

    @abstractmethod
    def correction(self) -> float:
        raise NotImplementedError


class GibbsHoltzOPC(OPC):
    """Overburden Pressure Correction according to ``Gibbs & Holtz (1957)``."""

    @property
    def eop(self) -> float:
        """Effective overburden pressure (:math:`kpa`)."""
        return self._eop

    @eop.setter
    @validators.le(280.0)
    @validators.gt(0.0)
    def eop(self, val: float) -> None:
        self._eop = val

    def correction(self) -> float:
        r"""SPT Correction.

        :Equation:

        .. math:: C_N = \dfrac{350}{\sigma_o + 70} \, \sigma_o \le 280kN/m^2

        :math:`\frac{N_c}{N_{60}}` should lie between 0.45 and 2.0, if
        :math:`\frac{N_c}{N_{60}}` is greater than 2.0, :math:`N_c` should be
        divided by 2.0 to obtain the design value used in finding the bearing
        capacity of the soil.
        """
        corr = 350.0 / (self.eop + 70.0)
        return corr / 2.0 if corr > 2.0 else corr


class BazaraaPeckOPC(OPC):
    """Overburden Pressure Correction according to ``Bazaraa (1967)``, and
    also by ``Peck and Bazaraa (1969)``.
    """

    #: Maximum effective overburden pressure (:math:`kPa`).
    STD_PRESSURE: Final = 71.8

    @property
    def eop(self) -> float:
        """Effective overburden pressure (:math:`kPa`)."""
        return self._eop

    @eop.setter
    @validators.ge(0.0)
    def eop(self, val: float) -> None:
        """Effective overburden pressure (:math:`kPa`)."""
        self._eop = val

    def correction(self) -> float:
        r"""SPT Correction.

        :Equation:

        .. math::

            C_N &= \dfrac{4}{1 + 0.0418 \cdot \sigma_o}, \, \sigma_o \lt 71.8kN/m^2

            C_N &= \dfrac{4}{3.25 + 0.0104 \cdot \sigma_o},
                   \, \sigma_o \gt 71.8kN/m^2

            C_N &= 1 \, , \, \sigma_o = 71.8kN/m^2
        """
        if isclose(self.eop, self.STD_PRESSURE, rel_tol=0.01):
            corr = 1.0
        elif self.eop < self.STD_PRESSURE:
            corr = 4.0 / (1.0 + 0.0418 * self.eop)
        else:
            corr = 4.0 / (3.25 + 0.0104 * self.eop)
        return corr


class PeckOPC(OPC):
    """Overburden Pressure Correction according to ``Peck et al. (1974)``."""

    @property
    def eop(self) -> float:
        """Effective overburden pressure (:math:`kPa`)."""
        return self._eop

    @eop.setter
    @validators.ge(24.0)
    def eop(self, val: float) -> None:
        self._eop = val

    def correction(self) -> float:
        r"""SPT Correction.

        :Equation:

        .. math:: C_N = 0.77 \log \left(\dfrac{2000}{\sigma_o} \right)
        """
        return 0.77 * log10(2000.0 / self.eop)


class LiaoWhitmanOPC(OPC):
    """Overburden Pressure Correction according to ``Liao & Whitman (1986)``.
    """

    @property
    def eop(self) -> float:
        """Effective overburden pressure (:math:`kPa`)."""
        return self._eop

    @eop.setter
    @validators.gt(0.0)
    def eop(self, val: float) -> None:
        self._eop = val

    def correction(self) -> float:
        r"""SPT Correction.

        :Equation:

        .. math:: C_N = \sqrt{\dfrac{100}{\sigma_o}}
        """
        return sqrt(100.0 / self.eop)


class SkemptonOPC(OPC):
    """Overburden Pressure Correction according to ``Skempton (1986)``."""

    @property
    def eop(self) -> float:
        """Effective overburden pressure (:math:`kPa`)."""
        return self._eop

    @eop.setter
    @validators.ge(0.0)
    def eop(self, val: float) -> None:
        self._eop = val

    def correction(self) -> float:
        r"""SPT Correction.

        :Equation:

        .. math:: C_N = \dfrac{2}{1 + 0.01044 \cdot \sigma_o}
        """
        return 2.0 / (1.0 + 0.01044 * self.eop)


class DilatancyCorrection:
    """Dilatancy SPT Correction according to ``Terzaghi & Peck (1948)``.

    For coarse sand, this correction is not required. In applying this
    correction, overburden pressure correction is applied first and then
    dilatancy correction is applied.
    """

    def __init__(self, corr_spt_n_value: float) -> None:
        """
        :param corr_spt_n_value: SPT N-value standardized for field procedures
                                and/or corrected for overburden pressure.
        :type corr_spt_n_value: float
        """
        self.corr_spt_n_value = corr_spt_n_value

    @property
    def corr_spt_n_value(self) -> float:
        """SPT N-value standardized for field procedures and/or corrected for
        overburden pressure.
        """
        return self._corr_spt_n_value

    @corr_spt_n_value.setter
    @validators.le(100.0)
    @validators.gt(0.0)
    def corr_spt_n_value(self, val: float) -> None:
        self._corr_spt_n_value = val

    @round_(ndigits=1)
    def corrected_spt_n_value(self) -> float:
        r"""Corrected SPT N-value.

        :Equation:

        .. math::

            (N_1)_{60} &= 15 + \dfrac{1}{2}((N_1)_{60} - 15) \, , \,
                          (N_1)_{60} \gt 15

            (N_1)_{60} &= (N_1)_{60} \, , \, (N_1)_{60} \le 15
        """
        if self.corr_spt_n_value <= 15.0:
            return self.corr_spt_n_value
        return 15.0 + 0.5 * (self.corr_spt_n_value - 15.0)


@enum_repr
class OPCType(enum.StrEnum):
    """Enumeration of overburden pressure correction types."""
    GIBBS = enum.auto()
    BAZARAA = enum.auto()
    PECK = enum.auto()
    LIAO = enum.auto()
    SKEMPTON = enum.auto()


_opctypes = {OPCType.GIBBS: GibbsHoltzOPC,
             OPCType.BAZARAA: BazaraaPeckOPC,
             OPCType.PECK: PeckOPC,
             OPCType.LIAO: LiaoWhitmanOPC,
             OPCType.SKEMPTON: SkemptonOPC}


def create_overburden_pressure_correction(std_spt_n_value: float, eop,
                                          opc_type: OPCType | str = OPCType.GIBBS) -> OPC:
    """A factory function that encapsulates the creation of overburden
    pressure correction.

    :param std_spt_n_value: SPT N-value standardized for field procedures.
    :type std_spt_n_value: float

    :param eop: Effective overburden pressure (:math:`kPa`).
    :type eop: float

    :param opc_type: Overburden Pressure Correction type to apply,
                    defaults to :py:enum:mem:`~OPCType.GIBBS`
    :type opc_type: OPCType, optional
    """

    try:
        opc_type = OPCType(str(opc_type).casefold())
    except ValueError as e:
        msg = ErrorMsg(param_name="opc_type",
                       param_value=opc_type,
                       symbol="in",
                       param_value_bound=list(OPCType))
        raise ValidationError(msg) from e

    opc_class = _opctypes[opc_type]
    opc_corr = opc_class(std_spt_n_value=std_spt_n_value, eop=eop)

    return opc_corr



import functools
import math
from math import exp, inf, isclose, log10, pi, sqrt
from statistics import fmean as mean
from typing import (Any, Callable, NotRequired, Optional, SupportsRound,
                    TypedDict, Unpack)

from . import validators

__all__ = ["ErrorMsg",
           "enum_repr",
           "inf",
           "pi",
           "deg2rad",
           "rad2deg",
           "tan",
           "cot",
           "sin",
           "cos",
           "arctan",
           "round_",
           "mean",
           "exp",
           "isclose",
           "log10",
           "sqrt",
           "validators"]


def deg2rad(x: float, /) -> float:
    """Convert angle x from degrees to radians."""
    return math.radians(x)


def rad2deg(x: float, /) -> float:
    """Convert angle x from radians to degrees."""
    return math.degrees(x)


def tan(x: float, /) -> float:
    """Return the tangent of x (measured in degrees)."""
    return math.tan(deg2rad(x))


def cot(x: float, /) -> float:
    """Return the cotangent of x (measured in degrees)."""
    return 1 / tan(x)


def sin(x: float, /) -> float:
    """Return the sine of x (measured in degrees)."""
    return math.sin(deg2rad(x))


def cos(x: float, /) -> float:
    """Return the cosine of x (measured in degrees)."""
    return math.cos(deg2rad(x))


def arctan(x: float, /) -> float:
    """Return the arc tangent (measured in degrees) of x."""
    return rad2deg(math.atan(x))


def enum_repr(cls):
    cls.__repr__ = lambda self: f"{self.value}"
    return cls


def round_(ndigits: int | Callable[..., SupportsRound]) -> Callable:
    """A decorator that rounds the result of a callable to a specified number
    of decimal places.

    The returned value of the callable should support the ``__round__`` dunder
    method and should be a numeric value. ``ndigits`` can either be an int
    which will indicate the number of decimal places to round to or a
    callable. If ``ndigits`` is callable the default decimal places is 2.

    TypeError is raised when ``ndigits`` is neither an int nor a callable.
    """

    def dec(fn) -> Callable[..., float]:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs) -> float:
            res = fn(*args, **kwargs)
            return round(res, ndigits=ndigits)

        return wrapper

    return dec



"""This module provides classes and utilities for modeling various foundation
footing types, their dimensions, and properties."""
import enum
from abc import ABC, abstractmethod
from typing import Optional, TypeVar

from .utils import enum_repr, inf, isclose, validators
from .utils.exceptions import ErrorMsg, ValidationError

__all__ = ["create_foundation",
           "FoundationSize",
           "FoundationType",
           "Shape",
           "StripFooting",
           "CircularFooting",
           "SquareFooting",
           "RectangularFooting"]

T = TypeVar("T")


@enum_repr
class Shape(enum.StrEnum):
    """Enumeration of foundation shapes."""
    STRIP = enum.auto()
    CIRCLE = enum.auto()
    SQUARE = enum.auto()
    RECTANGLE = enum.auto()


@enum_repr
class FoundationType(enum.StrEnum):
    """Enumeration of foundation types."""
    PAD = enum.auto()
    ISOLATED = PAD
    MAT = enum.auto()
    RAFT = MAT


class FootingSize(ABC):
    _SHAPE: Shape

    @property
    @abstractmethod
    def width(self) -> float:
        raise NotImplementedError

    @width.setter
    def width(self, value: float):
        raise NotImplementedError

    @property
    @abstractmethod
    def length(self) -> float:
        raise NotImplementedError

    @length.setter
    def length(self, value: float):
        raise NotImplementedError

    @property
    def shape(self) -> Shape:
        """Return the shape of the foundation footing."""
        return self._SHAPE


class StripFooting(FootingSize):
    """A class representation of strip footing."""

    _SHAPE = Shape.STRIP

    def __init__(self, width: float, length: float = inf) -> None:
        """
        :param width: Width of foundation footing (m).
        :type width: float

        :param float length: Length of foundation footing, defaults to inf (m).
        :type length: float
        """
        self.width = width
        self.length = length

    @property
    def width(self) -> float:
        """Width of foundation footing (m)."""
        return self._width

    @width.setter
    @validators.gt(0.0)
    def width(self, val: float):
        self._width = val

    @property
    def length(self) -> float:
        """Length of foundation footing (m)."""
        return self._length

    @length.setter
    @validators.ge(0.0)
    def length(self, val: float):
        self._length = val


class CircularFooting(FootingSize):
    """A class representation of circular footing.

    .. note::

        The ``width`` and ``length`` properties refer to the diameter of the
        circular footing. This is to make it compatible with the protocol
        square and rectangular footing follow.
    """

    _SHAPE = Shape.CIRCLE

    def __init__(self, diameter: float):
        """
        :param float diameter: Diameter of foundation footing (m).
        """
        self.diameter = diameter

    @property
    def diameter(self) -> float:
        """Diameter of foundation footing (m)."""
        return self._diameter

    @diameter.setter
    @validators.gt(0.0)
    def diameter(self, val: float):
        self._diameter = val

    @property
    def width(self):
        """Diameter of foundation footing (m)."""
        return self.diameter

    @width.setter
    def width(self, val: float):
        self.diameter = val

    @property
    def length(self):
        """Diameter of foundation footing (m)."""
        return self.diameter

    @length.setter
    def length(self, val: float):
        self.diameter = val


class SquareFooting(FootingSize):
    """A class representation of square footing."""

    _SHAPE = Shape.SQUARE

    def __init__(self, width: float):
        """
        :param float width: Width of foundation footing (m).
        """
        self.width = width

    @property
    def width(self):
        """Width of foundation footing (m)."""
        return self._width

    @width.setter
    @validators.gt(0)
    def width(self, val):
        self._width = val

    @property
    def length(self):
        """Width of foundation footing (m)."""
        return self.width

    @length.setter
    def length(self, val):
        self.width = val


class RectangularFooting(FootingSize):
    """A class representation of rectangular footing."""

    _SHAPE = Shape.RECTANGLE

    def __init__(self, width: float, length: float):
        """
        :param width: Width of foundation footing (m).
        :type width: float

        :param length: Length of foundation footing (m).
        :type length: float
        """
        self.width = width
        self.length = length

    @property
    def width(self) -> float:
        """Width of foundation footing (m)."""
        return self._width

    @width.setter
    @validators.gt(0.0)
    def width(self, val):
        self._width = val

    @property
    def length(self) -> float:
        """Length of foundation footing (m)."""
        return self._length

    @length.setter
    @validators.gt(0.0)
    def length(self, val):
        self._length = val


class FoundationSize:
    """A simple class representing a foundation structure."""

    def __init__(self, depth: float,
                 footing_size: FootingSize,
                 eccentricity: float = 0.0,
                 ground_water_level: Optional[float] = None,
                 foundation_type: FoundationType = FoundationType.PAD) -> None:
        """
        :param depth: Depth of foundation (m).
        :type depth: float

        :param footing_size: Represents the size of the foundation footing.
        :type footing_size: FootingSize

        :param eccentricity: The deviation of the foundation load from the
                             center of gravity of the foundation footing (m),
                             defaults to 0.0. This means that the foundation 
                             load aligns with the center of gravity of the 
                             foundation footing.
        :type eccentricity: float, optional

        :param ground_water_level: Depth of the water below ground level (m),
                                   defaults to None.
        :type ground_water_level: float, optional

        :param foundation_type: Type of foundation, defaults to
                                :py:enum:mem:`~FoundationType.PAD`
        :type foundation_type: FoundationType, optional
        """
        self.depth = depth
        self.footing_size = footing_size
        self.eccentricity = eccentricity

        self._ground_water_level = ground_water_level
        self._foundation_type = foundation_type

    @property
    def depth(self) -> float:
        """Depth of foundation (m)."""
        return self._depth

    @depth.setter
    @validators.gt(0.0)
    def depth(self, val: float) -> None:
        self._depth = val

    @property
    def width(self) -> float:
        """Width of foundation footing (m)."""
        return self.footing_size.width

    @width.setter
    def width(self, val: float):
        self.footing_size.width = val

    @property
    def length(self) -> float:
        """Length of foundation footing (m)."""
        return self.footing_size.length

    @length.setter
    def length(self, val: float):
        self.footing_size.length = val

    @property
    def footing_shape(self) -> Shape:
        """Shape of foundation footing."""
        return self.footing_size.shape

    @property
    def eccentricity(self) -> float:
        """The deviation of the foundation load from the center of gravity of
        the foundation footing (m).
        """
        return self._eccentricity

    @eccentricity.setter
    @validators.ge(0.0)
    def eccentricity(self, val: float) -> None:
        self._eccentricity = val

    @property
    def ground_water_level(self) -> Optional[float]:
        """Depth of the water below ground level (m)."""
        return self._ground_water_level

    @ground_water_level.setter
    @validators.ge(0.0)
    def ground_water_level(self, val: float) -> None:
        self._ground_water_level = val

    @property
    def foundation_type(self) -> FoundationType:
        """Type of foundation."""
        return self._foundation_type

    @property
    def effective_width(self) -> float:
        """Returns the effective width of the foundation footing (m)."""
        return self.width - 2.0 * self.eccentricity

    def footing_params(self) -> tuple[float, float, Shape]:
        """Returns the :attr:`effective_width`, :attr:`length`, and
        :attr:`footing_shape` of the foundation footing.
        """
        width, length, shape = (self.effective_width,
                                self.length,
                                self.footing_shape)

        if not isclose(width, length) and shape != Shape.STRIP:
            shape = Shape.RECTANGLE

        return width, length, shape


def create_foundation(depth: float,
                      width: float,
                      length: Optional[float] = None,
                      eccentricity: float = 0.0,
                      ground_water_level: Optional[float] = None,
                      foundation_type: FoundationType = FoundationType.PAD,
                      shape: Shape | str = Shape.SQUARE) -> FoundationSize:
    """A factory function that encapsulate the creation of a foundation.

    :param depth: Depth of foundation (m).
    :type depth: float

    :param width: Width of foundation footing. In the case of a circular
                  footing, it refers to the footing diameter (m).
    :type width: float

    :param length: Length of foundation footing (m), defaults to None.
    :type length: float, optional

    :param eccentricity: The deviation of the foundation load from the
                         center of gravity of the foundation footing (m),
                         defaults to 0.0. This means that the foundation
                         load aligns with the center of gravity of the
                         foundation footing .
    :type eccentricity: float, optional

    :param ground_water_level: Depth of the water below ground level (m),
                               defaults to None.
    :type ground_water_level: float, optional

    :param foundation_type: Type of foundation footing, defaults to
                            :py:enum:mem:`~FoundationType.PAD`.
    :type foundation_type: FoundationType, optional

    :param shape: Shape of foundation footing, defaults to 
                  :py:enum:mem:`~Shape.SQUARE`
    :type shape: Shape | str, optional

    :raises ValueError: Raised when length is not provided for a rectangular
                        footing.
    :raises ValueError: Raised if an invalid footing shape is provided.
    """

    try:
        shape = Shape(str(shape).casefold())
    except ValueError as e:
        msg = ErrorMsg(param_name="shape",
                       param_value=shape,
                       symbol="in",
                       param_value_bound=list(Shape))
        raise ValidationError(msg) from e

    if shape is Shape.STRIP:
        footing_size = StripFooting(width=width)
    elif shape is Shape.SQUARE:
        footing_size = SquareFooting(width=width)
    elif shape is Shape.CIRCLE:
        footing_size = CircularFooting(diameter=width)
    else:  # RECTANGLE
        if not length:
            msg = ErrorMsg(param_name="length",
                           param_value=length,
                           msg="Length of footing must be provided.")
            raise ValidationError(msg)
        footing_size = RectangularFooting(width=width, length=length)

    return FoundationSize(depth=depth,
                          eccentricity=eccentricity,
                          ground_water_level=ground_water_level,
                          foundation_type=foundation_type,
                          footing_size=footing_size)



from collections import UserString
from typing import Any


class _ErrorMsg(UserString):
    def __init__(self, param_name: str = None,
                 param_value: Any = None,
                 symbol: str = None,
                 param_value_bound: Any = None,
                 msg: str = None):
        if not msg:
            msg = f"{param_name}: {param_value!r} must be {symbol} {param_value_bound}"

        self.param_name = param_name
        self.param_value = param_value
        self.symbol = symbol
        self.param_value_bound = param_value_bound
        self.msg = msg

        super().__init__(msg)

    def __add__(self, other):
        if isinstance(other, str):
            self.data = self.data + other
            self.msg = self.data
            return self
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, str):
            self.data = other + self.data
            self.msg = self.data
            return self
        return NotImplemented


class ErrorMsg(_ErrorMsg):
    pass


class ValidationError(ValueError):
    """Exception raised when a validation error occurs."""

    def __init__(self, error: ErrorMsg):
        super().__init__(error)
        self.error = error


class SettlementError(ValidationError):
    """Raised when tolerable settlement is greater than the maximum
    allowable settlement.
    """



from geolysis.foundation import FoundationSize
from geolysis.utils import round_

from ._core import AllowableBearingCapacity


class MeyerhofABC4PadFoundation(AllowableBearingCapacity):
    r"""Allowable bearing capacity for pad foundation on cohesionless soils
    according to ``Meyerhof (1956)``.

    :Equation:

    .. math::

         q_a(kPa) &= 12N f_d\left(\dfrac{S}{25.4}\right), \ B \ \le 1.2m

         q_a(kPa) &= 8N\left(\dfrac{3.28B + 1}{3.28B} \right)^2 f_d\left(
                 \dfrac{S}{25.4}\right), \ B \ \gt 1.2m

         f_d &= 1 + 0.33 \cdot \frac{D_f}{B} \le 1.33

    ===================  ======================================  ===========
     Symbol                Description                              Unit
    ===================  ======================================  ===========
    :math:`q_a`          Allowable bearing capacity               :math:`kPa`
    :math:`N`            Corrected SPT N-value                     —
    :math:`f_d`          Depth factor                              —
    :math:`S`            Tolerable settlement                     :math:`mm`
    :math:`B`            Width of foundation footing              :math:`m`
    :math:`D_f`          Depth of foundation footing              :math:`m`
    ===================  ======================================  ===========
    """

    def __init__(self, corrected_spt_n_value: float,
                 tol_settlement: float,
                 foundation_size: FoundationSize):
        """
        :param corrected_spt_n_value: Average uncorrected SPT N-value (60% 
                                      energy with dilatancy (water) correction 
                                      if applicable) within the foundation 
                                      influence  zone i.e :math:`D_f` to
                                      :math:`D_f + 2B`.
        :type corrected_spt_n_value: float

        :param tol_settlement: Tolerable settlement of foundation (mm).
        :type tol_settlement: float

        :param foundation_size: Size of the foundation.
        :type foundation_size: FoundationSize
        """
        super().__init__(corrected_spt_n_value=corrected_spt_n_value,
                         tol_settlement=tol_settlement,
                         foundation_size=foundation_size)

    @round_(ndigits=2)
    def bearing_capacity(self):
        """Calculates the allowable bearing capacity of the pad foundation."""
        n_corr = self.corrected_spt_n_value
        width = self.foundation_size.width

        if width <= 1.2:
            return 12 * n_corr * self._fd() * self._sr()

        return (8 * n_corr * ((3.28 * width + 1) / (3.28 * width)) ** 2
                * self._fd() * self._sr())


class MeyerhofABC4MatFoundation(MeyerhofABC4PadFoundation):
    r"""Allowable bearing capacity for mat foundation on cohesionless soils
    according to ``Meyerhof (1956)``.

    :Equation:

    .. math::

         q_a(kPa) &= 8 N f_d\left(\dfrac{S}{25.4}\right)

         f_d &= 1 + 0.33 \cdot \frac{D_f}{B} \le 1.33

    ===================  ======================================  ===========
     Symbol                Description                              Unit
    ===================  ======================================  ===========
    :math:`q_a`          Allowable bearing capacity               :math:`kPa`
    :math:`N`            Corrected SPT N-value                     —
    :math:`f_d`          Depth factor                              —
    :math:`S`            Tolerable settlement                     :math:`mm`
    :math:`B`            Width of foundation footing              :math:`m`
    :math:`D_f`          Depth of foundation footing              :math:`m`
    ===================  ======================================  ===========
    """

    @round_(ndigits=2)
    def bearing_capacity(self):
        """Calculate the allowable bearing capacity of the mat foundation."""
        n_corr = self.corrected_spt_n_value
        return 8 * n_corr * self._fd() * self._sr()



"""This package provides a factory function and utilities for creating 
allowable bearing capacity calculations using methods like Bowles, Meyerhof, 
and Terzaghi for various foundation types and shapes.
"""
import enum
from typing import Optional

from geolysis.foundation import FoundationType, Shape, create_foundation
from geolysis.utils import enum_repr, inf
from geolysis.utils.exceptions import ErrorMsg, ValidationError

from ._core import AllowableBearingCapacity
from .bowles_abc import BowlesABC4MatFoundation, BowlesABC4PadFoundation
from .meyerhof_abc import MeyerhofABC4MatFoundation, MeyerhofABC4PadFoundation
from .terzaghi_abc import TerzaghiABC4MatFoundation, TerzaghiABC4PadFoundation


@enum_repr
class ABCType(enum.StrEnum):
    """Enumeration of available allowable bearing capacity types."""
    BOWLES = enum.auto()
    MEYERHOF = enum.auto()
    TERZAGHI = enum.auto()


abc_classes = {
    ABCType.BOWLES: {
        FoundationType.PAD: BowlesABC4PadFoundation,
        FoundationType.MAT: BowlesABC4MatFoundation,
    },
    ABCType.MEYERHOF: {
        FoundationType.PAD: MeyerhofABC4PadFoundation,
        FoundationType.MAT: MeyerhofABC4MatFoundation,
    },
    ABCType.TERZAGHI: {
        FoundationType.PAD: TerzaghiABC4PadFoundation,
        FoundationType.MAT: TerzaghiABC4MatFoundation,
    }
}


def create_allowable_bearing_capacity(corrected_spt_n_value: float,
                                      tol_settlement: float,
                                      depth: float,
                                      width: float,
                                      length: Optional[float] = None,
                                      eccentricity: float = 0.0,
                                      ground_water_level: float = inf,
                                      shape: Shape | str = Shape.SQUARE,
                                      foundation_type: FoundationType | str =
                                      FoundationType.PAD,
                                      abc_type: Optional[
                                          ABCType | str] = None,
                                      ) -> AllowableBearingCapacity:
    """ A factory function that encapsulate the creation of  allowable bearing
    capacities.

    :param corrected_spt_n_value: The corrected SPT N-value.
    :type corrected_spt_n_value: float

    :param tol_settlement: Tolerable settlement of foundation (mm).
    :type tol_settlement: float

    :param depth: Depth of foundation (m).
    :type depth: float

    :param width: Width of foundation footing (m).
    :type width: float

    :param length: Length of foundation footing (m).
    :type length: float, optional

    :param eccentricity: The deviation of the foundation load from the center 
                         of gravity of the foundation footing (m), defaults to
                         0.0. This means that the foundation load aligns with
                         the center of gravity of the foundation footing.
    :type eccentricity: float, optional

    :param ground_water_level: Depth of water below ground level (m).
    :type ground_water_level: float, optional

    :param shape: Shape of foundation footing, defaults to
                  :py:enum:mem:`~geolysis.foundation.Shape.SQUARE`.
    :type shape: str, optional

    :param foundation_type: Type of foundation, defaults to "pad".
    :type foundation_type: FoundationType | str, optional

    :param abc_type: Type of allowable bearing capacity calculation to apply.
                     Available values can be found in :py:enum:`ABCType`,
                     defaults to None.
    :type abc_type:  ABCType | str, optional

    :raises ValueError: Raised if ``abc_type`` or ``foundation_type`` is not
                        supported.
    :raises ValueError: Raised when ``length`` is not provided for a rectangular
                        footing.
    :raises ValueError: Raised if an invalid footing ``shape`` is provided.
    """

    msg = ErrorMsg(param_name="abc_type",
                   param_value=abc_type,
                   symbol="in",
                   param_value_bound=list(ABCType))

    if abc_type is None:
        raise ValidationError(msg)

    try:
        abc_type = ABCType(str(abc_type).casefold())
    except ValueError as e:
        raise ValidationError(msg) from e

    try:
        foundation_type = FoundationType(str(foundation_type).casefold())
    except ValueError as e:
        msg = ErrorMsg(param_name="foundation_type",
                       param_value=foundation_type,
                       symbol="in",
                       param_value_bound=list(FoundationType))
        raise ValidationError(msg) from e

    # exception from create_foundation will automaatically propagate
    # no need to catch and handle it.
    fnd_size = create_foundation(depth=depth,
                                 width=width,
                                 length=length,
                                 eccentricity=eccentricity,
                                 ground_water_level=ground_water_level,
                                 foundation_type=foundation_type,
                                 shape=shape)

    abc_class = abc_classes[abc_type][fnd_size.foundation_type]
    abc = abc_class(corrected_spt_n_value=corrected_spt_n_value,
                    tol_settlement=tol_settlement,
                    foundation_size=fnd_size)
    return abc



from geolysis.foundation import FoundationSize
from geolysis.utils import round_

from ._core import AllowableBearingCapacity


class TerzaghiABC4PadFoundation(AllowableBearingCapacity):
    r"""Allowable bearing capacity for pad foundation on cohesionless
    soils according to ``Terzaghi & Peck (1948)``.

    :Equation:

    .. math::

         q_a(kPa) &= 12N \dfrac{1}{c_w f_d}\left(\dfrac{S}{25.4}\right),
                     \ B \ \le 1.2m

         q_a(kPa) &= 8N\left(\dfrac{3.28B + 1}{3.28B} \right)^2\dfrac{1}
                     {c_w f_d}\left(\dfrac{S}{25.4}\right), \ B \ \gt 1.2m

         f_d &= 1 + 0.25 \cdot \frac{D_f}{B} \le 1.25

         c_w &= 2 - \frac{D_w}{2B} \le 2, D_w \gt D_f

         c_w &= 2 - \frac{D_f}{2B} \le 2, D_w \le D_f

    ===================  ======================================  ===========
     Symbol                Description                              Unit
    ===================  ======================================  ===========
    :math:`q_a`          Allowable bearing capacity               :math:`kPa`
    :math:`N`            Corrected SPT N-value                     —
    :math:`f_d`          Depth factor                              —
    :math:`c_w`          Water correction factor                   —
    :math:`S`            Tolerable settlement                     :math:`mm`
    :math:`B`            Width of foundation footing              :math:`m`
    :math:`D_f`          Depth of foundation footing              :math:`m`
    :math:`D_w`          Depth of water below ground level        :math:`m`
    ===================  ======================================  ===========
    """

    def __init__(self, corrected_spt_n_value: float,
                 tol_settlement: float,
                 foundation_size: FoundationSize) -> None:
        """
        :param corrected_spt_n_value: Lowest (or average) uncorrected SPT 
                                      N-value (60% energy) within the foundation 
                                      influence zone i.e :math:`D_f` to 
                                      :math:`D_f + 2B`
        :type corrected_spt_n_value: float

        :param tol_settlement: Tolerable settlement of foundation (mm).
        :type tol_settlement: float

        :param foundation_size: Size of the foundation.
        :type foundation_size: FoundationSize
        """
        super().__init__(corrected_spt_n_value=corrected_spt_n_value,
                         tol_settlement=tol_settlement,
                         foundation_size=foundation_size)

    def _fd(self) -> float:
        """Calculate the depth factor."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width

        return min(1.0 + 0.25 * depth / width, 1.25)

    def _cw(self):
        """Calculate the water correction factor."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        water_level = self.foundation_size.ground_water_level

        if water_level is None:
            return 2.0

        if water_level <= depth:
            cw = 2.0 - depth / (2.0 * width)
        else:
            cw = 2.0 - water_level / (2.0 * width)

        return min(cw, 2.0)

    @round_(ndigits=2)
    def bearing_capacity(self):
        """Calculates the allowable bearing capacity of the pad foundation."""
        n_corr = self.corrected_spt_n_value
        width = self.foundation_size.width

        if width <= 1.2:
            return 12 * n_corr * (1 / (self._cw() * self._fd())) * self._sr()

        return (8 * n_corr * ((3.28 * width + 1) / (3.28 * width)) ** 2
                * (1 / (self._cw() * self._fd())) * self._sr())


class TerzaghiABC4MatFoundation(TerzaghiABC4PadFoundation):
    r"""Allowable bearing capacity for mat foundation on cohesionless soils
    according to ``Terzaghi & Peck (1948)``.

    :Equation:

    .. math::

         q_a(kPa) &= 8N\dfrac{1}{c_w f_d}\left(\dfrac{S}{25.4}\right)

         f_d &= 1 + 0.25 \cdot \frac{D_f}{B} \le 1.25

         c_w &= 2 - \frac{D_w}{2B} \le 2, D_w \gt D_f

         c_w &= 2 - \frac{D_f}{2B} \le 2, D_w \le D_f

    ===================  ======================================  ===========
     Symbol                Description                              Unit
    ===================  ======================================  ===========
    :math:`q_a`          Allowable bearing capacity               :math:`kPa`
    :math:`N`            Corrected SPT N-value                     —
    :math:`f_d`          Depth factor                              —
    :math:`c_w`          Water correction factor                   —
    :math:`S`            Tolerable settlement                     :math:`mm`
    :math:`B`            Width of foundation footing              :math:`m`
    :math:`D_f`          Depth of foundation footing              :math:`m`
    :math:`D_w`          Depth of water below ground level        :math:`m`
    ===================  ======================================  ===========
    """

    @round_(ndigits=2)
    def bearing_capacity(self):
        """Calculates the allowable bearing capacity of the mat foundation."""
        n_corr = self.corrected_spt_n_value
        return 8 * n_corr * (1 / (self._cw() * self._fd())) * self._sr()



import pytest

from geolysis.spt import (DilatancyCorrection, EnergyCorrection, HammerType,
                          SamplerType, SPTNDesign,
                          create_overburden_pressure_correction)
from geolysis.utils.exceptions import ValidationError


def test_create_spt_correction_errors():
    with pytest.raises(ValidationError):
        create_overburden_pressure_correction(std_spt_n_value=34,
                                              eop=100, opc_type="TERZAGHI")


class TestSPTNDesign:

    def test_spt_n_design(self):
        spt_design = SPTNDesign([7.0, 15.0, 18], method="min")

        assert spt_design.n_design() == pytest.approx(7.0)
        spt_design.method = "avg"
        assert spt_design.n_design() == pytest.approx(13.3)
        spt_design.method = "wgt"
        assert spt_design.n_design() == pytest.approx(9.4)

    def test_errors(self):
        # Provided an empty value for corrected_spt_n_values
        with pytest.raises(ValidationError):
            SPTNDesign(corrected_spt_n_values=[])

        # Provided an invalid method
        with pytest.raises(ValidationError):
            SPTNDesign(corrected_spt_n_values=[7.0, 15.0, 18], method="max")

        # corrected_spt_n_values is greater than 100
        with pytest.raises(ValidationError):
            SPTNDesign(corrected_spt_n_values=[22, 44, 120])

        # corrected_spt_n_values is 0.0
        with pytest.raises(ValidationError):
            SPTNDesign(corrected_spt_n_values=[0.0, 15.0, 18])

        # corrected_spt_n_values is less than 0.0
        with pytest.raises(ValidationError):
            SPTNDesign(corrected_spt_n_values=[-10, 15.0, 18])


class TestEnergyCorrection:

    @pytest.mark.parametrize(
        ["rec_spt_n_val", "energy_percentage", "borehole_diameter",
         "rod_len", "hammer_type", "sampler_type", "expected"],
        [(30, 0.6, 65.0, 3, HammerType.DONUT_1, SamplerType.STANDARD, 22.5),
         (30, 0.6, 120.0, 5, HammerType.AUTOMATIC, SamplerType.NON_STANDARD,
          37.5),
         (30, 0.6, 170.0, 7, HammerType.DONUT_2, SamplerType.STANDARD, 27.3),
         (30, 0.6, 65.0, 12, HammerType.SAFETY, SamplerType.STANDARD, 27.5)])
    def test_energy_correction(self, rec_spt_n_val,
                               energy_percentage,
                               borehole_diameter,
                               rod_len,
                               hammer_type,
                               sampler_type,
                               expected):
        energy_corr = EnergyCorrection(recorded_spt_n_value=rec_spt_n_val,
                                       energy_percentage=energy_percentage,
                                       borehole_diameter=borehole_diameter,
                                       rod_length=rod_len,
                                       hammer_type=hammer_type,
                                       sampler_type=sampler_type)

        assert energy_corr.standardized_spt_n_value() == pytest.approx(
            expected)

    def test_errors(self):
        # Provided an invalid value for hammer_type
        with pytest.raises(ValidationError):
            EnergyCorrection(recorded_spt_n_value=22, hammer_type="manual")

        # Provided an invalid value for sampler_type
        with pytest.raises(ValidationError):
            EnergyCorrection(recorded_spt_n_value=22, sampler_type="std")


class TestGibbsHoltzOPC:

    @pytest.mark.parametrize(["std_spt_n_value", "eop", "expected"],
                             [(22.5, 100.0, 23.2)])
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="GIBBS")
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestBazaraaPeckOPC:

    @pytest.mark.parametrize(["std_spt_n_value", "eop", "expected"],
                             [(11.4, 54.8, 13.9),
                              (22.5, 100.0, 21.0),
                              (22.5, 71.8, 22.5)])
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="BAZARAA")
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestPeckOPC:

    @pytest.mark.parametrize(["std_spt_n_value", "eop", "expected"],
                             [(22.5, 100.0, 22.5)])
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="PECK")
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestLiaoWhitmanOPC:

    @pytest.mark.parametrize(["std_spt_n_value", "eop", "expected"],
                             [(22.5, 100.0, 22.5)])
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="LIAO")
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestSkemptonOPC:
    @pytest.mark.parametrize(["std_spt_n_value", "eop", "expected"],
                             [(22.5, 100.0, 22.0)])
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="SKEMPTON")
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestDilatancyCorrection:

    @pytest.mark.parametrize(["std_spt_n_value", "expected"],
                             [(22.5, 18.8), (12.6, 12.6)])
    def test_correction(self, std_spt_n_value, expected):
        corr = DilatancyCorrection(corr_spt_n_value=std_spt_n_value)
        assert corr.corrected_spt_n_value() == pytest.approx(expected)



"""This module provides classes for soil classification using systems like 
USCS and AASHTO, based on particle size distribution and  Atterberg limits.
"""
import enum
from typing import NamedTuple, Optional, Sequence

from .utils import enum_repr, isclose, round_, validators
from .utils.exceptions import ErrorMsg, ValidationError

__all__ = ["ClfType",
           "AtterbergLimits",
           "PSD",
           "AASHTO",
           "USCS",
           "create_soil_classifier"]


class SizeDistError(ZeroDivisionError):
    """Exception raised when size distribution is not provided."""


@enum_repr
class _Clf(tuple, enum.Enum):

    def __str__(self) -> str:
        return self.name

    @property
    def symbol(self) -> str:
        return self.value[0]

    @property
    def description(self) -> str:
        return self.value[1]


class USCSSymbol(_Clf):
    """Unified Soil Classification System (USCS) symbols and descriptions."""
    G = GRAVEL = ("G", "Gravel")
    S = SAND = ("S", "Sand")
    M = SILT = ("M", "Silt")
    C = CLAY = ("C", "Clay")
    O = ORGANIC = ("O", "Organic")
    W = WELL_GRADED = ("W", "Well graded")
    P = POORLY_GRADED = ("P", "Poorly graded")
    L = LOW_PLASTICITY = ("L", "Low plasticity")
    H = HIGH_PLASTICITY = ("H", "High plasticity")
    GW = ("GW", "Well graded gravels")
    GP = ("GP", "Poorly graded gravels")
    GM = ("GM", "Silty gravels")
    GC = ("GC", "Clayey gravels")
    GM_GC = ("GM-GC", "Gravelly clayey silt")
    GW_GM = ("GW-GM", "Well graded gravel with silt")
    GP_GM = ("GP-GM", "Poorly graded gravel with silt")
    GW_GC = ("GW-GC", "Well graded gravel with clay")
    GP_GC = ("GP-GC", "Poorly graded gravel with clay")
    SW = ("SW", "Well graded sands")
    SP = ("SP", "Poorly graded sands")
    SM = ("SM", "Silty sands")
    SC = ("SC", "Clayey sands")
    SM_SC = ("SM-SC", "Sandy clayey silt")
    SW_SM = ("SW-SM", "Well graded sand with silt")
    SP_SM = ("SP-SM", "Poorly graded sand with silt")
    SW_SC = ("SW-SC", "Well graded sand with clay")
    SP_SC = ("SP-SC", "Poorly graded sand with clay")
    ML = ("ML", "Inorganic silts with low plasticity")
    CL = ("CL", "Inorganic clays with low plasticity")
    ML_CL = ("ML-CL", "Clayey silt with low plasticity")
    OL = ("OL", "Organic clays with low plasticity")
    MH = ("MH", "Inorganic silts with high plasticity")
    CH = ("CH", "Inorganic clays with high plasticity")
    OH = ("OH", "Organic silts with high plasticity")
    Pt = ("Pt", "Highly organic soils")


class AASHTOSymbol(_Clf):
    """AASHTO soil classification symbols and descriptions."""
    A_1_a = ("A-1-a", "Stone fragments, gravel, and sand")
    A_1_b = ("A-1-b", "Stone fragments, gravel, and sand")
    A_3 = ("A-3", "Fine sand")
    A_2_4 = ("A-2-4", "Silty or clayey gravel and sand")
    A_2_5 = ("A-2-5", "Silty or clayey gravel and sand")
    A_2_6 = ("A-2-6", "Silty or clayey gravel and sand")
    A_2_7 = ("A-2-7", "Silty or clayey gravel and sand")
    A_4 = ("A-4", "Silty soils")
    A_5 = ("A-5", "Silty soils")
    A_6 = ("A-6", "Clayey soils")
    A_7_5 = ("A-7-5", "Clayey soils")
    A_7_6 = ("A-7-6", "Clayey soils")


class AtterbergLimits:
    """Represents the water contents at which soil changes from one state to 
    the other.
    """

    class __A_LINE:

        def __get__(self, obj, objtype=None) -> float:
            return 0.73 * (obj.liquid_limit - 20.0)

    #:  The ``A-line`` determines if a soil is clayey or silty.
    #:  :math:`A = 0.73(LL - 20.0)`
    A_LINE = __A_LINE()

    def __init__(self, liquid_limit: float, plastic_limit: float):
        """
        :param liquid_limit: Water content beyond which soils flows under their 
                             own weight (%). It can also be defined as the
                             minimum moisture content at which a soil flows upon
                             application of a very small shear force.
        :type liquid_limit: float

        :param plastic_limit: Water content at which plastic deformation can be 
                              initiated (%). It is also the minimum water
                              content at which soil can be rolled into a thread
                              3mm thick. (molded without breaking)
        :type plastic_limit: float
        """
        if liquid_limit < plastic_limit:
            raise ValueError("liquid_limit cannot be less than plastic_limit")

        self.liquid_limit = liquid_limit
        self.plastic_limit = plastic_limit

    @property
    def liquid_limit(self) -> float:
        """Water content beyond which soils flows under their own weight (%)."""
        return self._liquid_limit

    @liquid_limit.setter
    @validators.ge(0.0)
    def liquid_limit(self, val: float) -> None:
        self._liquid_limit = val

    @property
    def plastic_limit(self) -> float:
        """Water content at which plastic deformation can be initiated (%)."""
        return self._plastic_limit

    @plastic_limit.setter
    @validators.ge(0.0)
    def plastic_limit(self, val: float) -> None:
        self._plastic_limit = val

    @property
    @round_(2)
    def plasticity_index(self) -> float:
        """Plasticity index (PI) is the range of water content over which the
        soil remains in the plastic state.

        It is also the numerical difference between the liquid limit and
        plastic limit of the soil.

        :Equation:

        .. math:: PI = LL - PL
        """
        return self.liquid_limit - self.plastic_limit

    @property
    def fine_material_type(self) -> USCSSymbol:
        """Checks whether the soil is either clay or silt."""
        return USCSSymbol.CLAY if self.above_A_LINE() else USCSSymbol.SILT

    def above_A_LINE(self) -> bool:
        """Checks if the soil sample is above A-Line."""
        return self.plasticity_index > self.A_LINE

    def limit_plot_in_hatched_zone(self) -> bool:
        """Checks if soil sample plot in the hatched zone on the atterberg
        chart.
        """
        return 4 <= self.plasticity_index <= 7 and 10 < self.liquid_limit < 30

    @round_(ndigits=2)
    def liquidity_index(self, nmc: float) -> float:
        r"""Return the liquidity index of the soil.

        Liquidity index of a soil indicates the nearness of its ``natural water
        content`` to its ``liquid limit``. When the soil is at the plastic
        limit its liquidity index is zero. Negative values of the liquidity
        index indicate that the soil is in a hard (desiccated) state. It is
        also known as Water-Plasticity ratio.

        :param float nmc: Moisture contents of the soil in natural condition.

        :Equation:

        .. math:: I_l = \dfrac{w - PL}{PI} \cdot 100
        """
        return ((nmc - self.plastic_limit) / self.plasticity_index) * 100.0

    @round_(2)
    def consistency_index(self, nmc: float) -> float:
        r"""Return the consistency index of the soil.

        Consistency index indicates the consistency (firmness) of soil. It
        shows the nearness of the ``natural water content`` of the soil to its
        ``plastic limit``. When the soil is at the liquid limit, the
        consistency index is zero. The soil at consistency index of zero will
        be extremely soft and has negligible shear strength. A soil at a water
        content equal to the plastic limit has consistency index of 100%
        indicating that the soil is relatively firm. A consistency index of
        greater than 100% shows the soil is relatively strong
        (semi-solid state). A negative value indicate the soil is in the liquid
        state. It is also known as Relative Consistency.

        :param float nmc: Moisture contents of the soil in natural condition.

        :Equation:

        .. math:: I_c = \dfrac{LL - w}{PI} \cdot 100
        """
        return ((self.liquid_limit - nmc) / self.plasticity_index) * 100.0


class _SizeDistribution:
    """Particle size distribution of soil sample.

    Features obtained from the Particle Size Distribution graph.
    """

    def __init__(self, d_10: float = 0.0,
                 d_30: float = 0.0,
                 d_60: float = 0.0):
        self.d_10 = d_10
        self.d_30 = d_30
        self.d_60 = d_60

    def __iter__(self):
        return iter([self.d_10, self.d_30, self.d_60])

    @property
    def coeff_of_curvature(self) -> float:
        return (self.d_30 ** 2.0) / (self.d_60 * self.d_10)

    @property
    def coeff_of_uniformity(self) -> float:
        return self.d_60 / self.d_10

    def grade(self, coarse_soil: USCSSymbol) -> USCSSymbol:
        """Grade of soil sample. Soil grade can either be well graded or poorly
        graded.

        :param coarse_soil: Coarse fraction of the soil sample. Valid arguments 
                            are :py:enum:mem:`~USCSSymbol.GRAVEL` and
                            :py:enum:mem:`~USCSSymbol.SAND`.
        """
        if coarse_soil is USCSSymbol.GRAVEL:
            if 1 < self.coeff_of_curvature < 3 and self.coeff_of_uniformity >= 4:
                grade = USCSSymbol.WELL_GRADED
            else:
                grade = USCSSymbol.POORLY_GRADED
            return grade

        # coarse soil is sand
        if 1 < self.coeff_of_curvature < 3 and self.coeff_of_uniformity >= 6:
            grade = USCSSymbol.WELL_GRADED
        else:
            grade = USCSSymbol.POORLY_GRADED
        return grade


class PSD:
    """Quantitative proportions by mass of various sizes of particles present
    in a soil.
    """

    def __init__(self, fines: float, sand: float,
                 d_10: float = 0, d_30: float = 0, d_60: float = 0):
        """
        :param fines: Percentage of fines in soil sample (%) i.e. The
                      percentage of soil sample passing through No. 200 sieve
                      (0.075mm).
        :type fines: float

        :param sand: Percentage of sand in soil sample (%).
        :type sand: float

        :param d_10: Diameter at which 10% of the soil by weight is finer.
        :type d_10: float
        :param d_30: Diameter at which 30% of the soil by weight is finer.
        :type d_30: float
        :param d_60: Diameter at which 60% of the soil by weight is finer.
        :type d_60: float
        """
        self.fines = fines
        self.sand = sand
        self.size_dist = _SizeDistribution(d_10=d_10, d_30=d_30, d_60=d_60)

    @property
    def gravel(self):
        """Percentage of gravel in soil sample (%)."""
        return 100.0 - (self.fines + self.sand)

    @property
    def coarse_material_type(self) -> USCSSymbol:
        """Determines whether the soil is either gravel or sand."""
        if self.gravel > self.sand:
            return USCSSymbol.GRAVEL
        return USCSSymbol.SAND

    @property
    @round_(ndigits=2)
    def coeff_of_curvature(self) -> float:
        r"""Coefficient of curvature of soil sample.

        :Equation:

        .. math:: C_c = \dfrac{D^2_{30}}{D_{60} \cdot D_{10}}

        For the soil to be well graded, the value of :math:`C_c` must be
        between 1 and 3.
        """
        return self.size_dist.coeff_of_curvature

    @property
    @round_(ndigits=2)
    def coeff_of_uniformity(self) -> float:
        r"""Coefficient of uniformity of soil sample.

        :Equation:

        .. math:: C_u = \dfrac{D_{60}}{D_{10}}

        :math:`C_u` value greater than 4 to 6 classifies the soil as well
        graded for gravels and sands respectively. When :math:`C_u` is less
        than 4, it is classified as poorly graded or uniformly graded soil.

        Higher values of :math:`C_u` indicates that the soil mass consists of
        soil particles with different size ranges.
        """
        return self.size_dist.coeff_of_uniformity

    def has_particle_sizes(self) -> bool:
        """Checks if soil sample has particle sizes."""
        return any(self.size_dist)

    def grade(self) -> USCSSymbol:
        r"""Return the grade of the soil sample, either well graded or poorly
        graded.

        Conditions for a well-graded soil:

        :Equation:

        - :math:`1 \lt C_c \lt 3` and :math:`C_u \ge 4` (for gravels)
        - :math:`1 \lt C_c \lt 3` and :math:`C_u \ge 6` (for sands)
        """
        return self.size_dist.grade(coarse_soil=self.coarse_material_type)


class SoilClf(NamedTuple):
    symbol: str
    description: str


class AASHTO:
    r"""American Association of State Highway and Transportation Officials
    (AASHTO) classification system.

    The AASHTO classification system is useful for classifying soils for
    highways. It categorizes soils for highways based on particle size analysis
    and plasticity characteristics. It classifies both coarse-grained and
    fine-grained soils into eight main groups (A1-A7) with subgroups, along with
    a separate category (A8) for organic soils.

    - ``A1 ~ A3`` (Granular Materials) :math:`\le` 35% pass No. 200 sieve
    - ``A4 ~ A7`` (Silt-clay Materials) :math:`\ge` 36% pass No. 200 sieve
    - ``A8`` (Organic Materials)

    The Group Index ``(GI)`` is used to further evaluate soils within a group.

    .. note::

        The ``GI`` must be mentioned even when it is zero, to indicate that the 
        soil has been classified as per AASHTO system.

    :Equation:

    .. math::

        GI = (F_{200} - 35)[0.2 + 0.005(LL - 40)] + 0.01(F_{200} - 15)(PI - 10)
    """

    def __init__(self, atterberg_limits: AtterbergLimits,
                 fines: float, add_group_idx: bool = True):
        """
        :param atterberg_limits: Atterberg limits of soil sample.
        :type atterberg_limits: AtterbergLimits

        :param fines: Percentage of fines in soil sample (%) i.e. The percentage
                      of soil sample passing through No. 200 sieve (0.075mm).
        :type fines: float

        :param add_group_idx: Used to indicate whether the group index should
                              be added to the classification or not, defaults
                              to True.
        :type add_group_idx: bool, optional
        """
        self.atterberg_limits = atterberg_limits
        self.fines = fines
        self.add_group_idx = add_group_idx

    @property
    def fines(self) -> float:
        """Percentage of fines in soil sample (%)."""
        return self._fines

    @fines.setter
    @validators.ge(0.0)
    def fines(self, val: float) -> None:
        self._fines = val

    @round_(ndigits=0)
    def group_index(self) -> float:
        """Return the Group Index (GI) of the soil sample."""

        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index
        fines = self.fines

        var_a = 1.0 if (x_0 := fines - 35.0) < 0.0 else min(x_0, 40.0)
        var_b = 1.0 if (x_0 := liquid_lmt - 40.0) < 0.0 else min(x_0, 20.0)
        var_c = 1.0 if (x_0 := fines - 15.0) < 0.0 else min(x_0, 40.0)
        var_d = 1.0 if (x_0 := plasticity_idx - 10.0) < 0.0 else min(x_0, 20.0)

        return var_a * (0.2 + 0.005 * var_b) + 0.01 * var_c * var_d

    def classify(self) -> SoilClf:
        """Return the AASHTO classification of the soil."""
        soil_clf = self._classify()

        symbol, description = soil_clf.symbol, soil_clf.description

        if self.add_group_idx:
            symbol = f"{symbol}({self.group_index():.0f})"

        return SoilClf(symbol, description)

    def _classify(self) -> AASHTOSymbol:
        # Silts A4-A7
        if self.fines > 35:
            soil_clf = self._fine_soil_classifier()
        # Coarse A1-A3
        else:
            soil_clf = self._coarse_soil_classifier()

        return soil_clf

    def _fine_soil_classifier(self) -> AASHTOSymbol:
        # A-4 -> A-5, Silty Soils
        # A-6 -> A-7, Clayey Soils
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index

        if liquid_lmt <= 40:
            if plasticity_idx <= 10.0:
                soil_clf = AASHTOSymbol.A_4
            else:
                soil_clf = AASHTOSymbol.A_6
        else:
            if plasticity_idx <= 10.0:
                soil_clf = AASHTOSymbol.A_5
            else:
                if plasticity_idx <= (liquid_lmt - 30.0):
                    soil_clf = AASHTOSymbol.A_7_5
                else:
                    soil_clf = AASHTOSymbol.A_7_6

        return soil_clf

    def _coarse_soil_classifier(self) -> AASHTOSymbol:
        # A-3, Fine sand
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index

        if self.fines <= 10.0 and isclose(plasticity_idx, 0.0, rel_tol=0.01):
            soil_clf = AASHTOSymbol.A_3
        # A-1-a -> A-1-b, Stone fragments, gravel, and sand
        elif self.fines <= 15 and plasticity_idx <= 6:
            soil_clf = AASHTOSymbol.A_1_a
        elif self.fines <= 25 and plasticity_idx <= 6:
            soil_clf = AASHTOSymbol.A_1_b
        # A-2-4 -> A-2-7, Silty or clayey gravel and sand
        elif liquid_lmt <= 40:
            if plasticity_idx <= 10:
                soil_clf = AASHTOSymbol.A_2_4
            else:
                soil_clf = AASHTOSymbol.A_2_6
        else:
            if plasticity_idx <= 10:
                soil_clf = AASHTOSymbol.A_2_5
            else:
                soil_clf = AASHTOSymbol.A_2_7

        return soil_clf


class USCS:
    """Unified Soil Classification System (USCS).

    The Unified Soil Classification System, initially developed by Casagrande
    in 1948 and later modified in 1952, is widely utilized in engineering
    projects involving soils. It is the most popular system for soil
    classification and is similar to Casagrande's Classification System. The
    system relies on particle size analysis and atterberg limits for
    classification.

    In this system, soils are first classified into two categories:

    - Coarse grained soils: If more than 50% of the soils is retained on
      No. 200 (0.075 mm) sieve, it is designated as coarse-grained soil.

    - Fine grained soils: If more than 50% of the soil passes through No. 200
      sieve, it is designated as fine-grained soil.

    Highly Organic soils are identified by visual inspection. These soils are
    termed as Peat. (:math:`P_t`)
    """

    def __init__(self, atterberg_limits: AtterbergLimits,
                 psd: PSD, organic=False):
        """
        :param atterberg_limits: Atterberg limits of the soil.
        :type atterberg_limits: AtterbergLimits

        :param psd: Particle size distribution of the soil.
        :type psd: PSD

        :param organic: Indicates whether soil is organic or not, defaults to 
                        False.
        :type organic: bool, optional
        """
        self.atterberg_limits = atterberg_limits
        self.psd = psd
        self.organic = organic

    def classify(self) -> SoilClf:
        """Return the USCS classification of the soil."""
        soil_clf = self._classify()

        # Ensure soil_clf is of type USCSSymbol
        if isinstance(soil_clf, str):
            soil_clf = USCSSymbol[soil_clf]

        if isinstance(soil_clf, USCSSymbol):
            return SoilClf(soil_clf.symbol, soil_clf.description)

        # Handling tuple or list case for dual classification
        first_clf, second_clf = map(lambda clf: USCSSymbol[clf], soil_clf)

        comb_symbol = f"{first_clf.symbol},{second_clf.symbol}"
        comb_desc = f"{first_clf.description},{second_clf.description}"

        return SoilClf(comb_symbol, comb_desc)

    def _classify(self) -> USCSSymbol | str | Sequence[str]:
        # Fine-grained, Run Atterberg
        if self.psd.fines > 50.0:
            return self._fine_soil_classifier()
        # Coarse grained, Run Sieve Analysis
        # Gravel or Sand
        return self._coarse_soil_classifier()

    def _fine_soil_classifier(self) -> USCSSymbol:
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index

        if liquid_lmt < 50.0:
            # Low LL
            # Above A-line and PI > 7
            if self.atterberg_limits.above_A_LINE() and plasticity_idx > 7.0:
                soil_clf = USCSSymbol.CL

            # Limit plot in hatched area on plasticity chart
            elif self.atterberg_limits.limit_plot_in_hatched_zone():
                soil_clf = USCSSymbol.ML_CL

            # Below A-line or PI < 4
            else:
                soil_clf = USCSSymbol.OL if self.organic else USCSSymbol.ML

        # High LL
        else:
            # Above A-Line
            if self.atterberg_limits.above_A_LINE():
                soil_clf = USCSSymbol.CH

            # Below A-Line
            else:
                soil_clf = USCSSymbol.OH if self.organic else USCSSymbol.MH

        return soil_clf

    def _coarse_soil_classifier(self) -> USCSSymbol | str | Sequence[str]:
        coarse_material_type = self.psd.coarse_material_type

        # More than 12% pass No. 200 sieve
        if self.psd.fines > 12.0:
            # Above A-line
            if self.atterberg_limits.above_A_LINE():
                soil_clf = f"{coarse_material_type}{USCSSymbol.CLAY}"

            # Limit plot in hatched zone on plasticity chart
            elif self.atterberg_limits.limit_plot_in_hatched_zone():
                if coarse_material_type == USCSSymbol.GRAVEL:
                    soil_clf = USCSSymbol.GM_GC
                else:
                    soil_clf = USCSSymbol.SM_SC

            # Below A-line
            else:
                if coarse_material_type == USCSSymbol.GRAVEL:
                    soil_clf = USCSSymbol.GM
                else:
                    soil_clf = USCSSymbol.SM

        elif 5.0 <= self.psd.fines <= 12.0:
            # Requires dual symbol based on graduation and plasticity chart
            if self.psd.has_particle_sizes():
                soil_clf = self._dual_soil_classifier()
            else:
                fine_material_type = self.atterberg_limits.fine_material_type

                soil_clf = (f"{coarse_material_type}{USCSSymbol.WELL_GRADED}_"
                            f"{coarse_material_type}{fine_material_type}",
                            f"{coarse_material_type}{USCSSymbol.POORLY_GRADED}_"
                            f"{coarse_material_type}{fine_material_type}")

        # Less than 5% pass No. 200 sieve
        # Obtain Cc and Cu from grain size graph
        else:
            if self.psd.has_particle_sizes():
                soil_clf = f"{coarse_material_type}{self.psd.grade()}"

            else:
                soil_clf = (f"{coarse_material_type}{USCSSymbol.WELL_GRADED}",
                            f"{coarse_material_type}{USCSSymbol.POORLY_GRADED}")

        return soil_clf

    def _dual_soil_classifier(self) -> str:
        fine_material_type = self.atterberg_limits.fine_material_type
        coarse_material_type = self.psd.coarse_material_type

        return (f"{coarse_material_type}{self.psd.grade()}_"
                f"{coarse_material_type}{fine_material_type}")


@enum_repr
class ClfType(enum.StrEnum):
    """Enumeration of soil classification types."""
    AASHTO = enum.auto()
    USCS = enum.auto()


def create_soil_classifier(liquid_limit: float,
                           plastic_limit: float,
                           fines: float,
                           sand: Optional[float] = None,
                           d_10: float = 0, d_30: float = 0, d_60: float = 0,
                           add_group_idx: bool = True,
                           organic: bool = False,
                           clf_type: Optional[ClfType | str] = None
                           ) -> AASHTO | USCS:
    """ A factory function that encapsulates the creation of a soil classifier.

    :param liquid_limit: Water content beyond which soils flows under their own
                         weight (%). It can also be defined as the minimum
                         moisture content at which a soil flows upon application
                         of a very small shear force.
    :type liquid_limit: float

    :param plastic_limit: Water content at which plastic deformation can be
                          initiated (%). It is also the minimum water content at
                          which soil can be rolled into a thread 3mm thick.
                          (molded without breaking)
    :type plastic_limit: float

    :param fines: Percentage of fines in soil sample (%) i.e. The percentage of
                  soil sample passing through No. 200 sieve (0.075mm).
    :type fines: float

    :param sand: Percentage of sand in soil sample (%). This is optional for
                 :class:`AASHTO` classification.
    :type sand: float, optional

    :param d_10: Diameter at which 10% of the soil by weight is finer.
    :type d_10: float, optional

    :param d_30: Diameter at which 30% of the soil by weight is finer.
    :type d_30: float, optional

    :param d_60: Diameter at which 60% of the soil by weight is finer.
    :type d_60: float, optional

    :param add_group_idx: Used to indicate whether the group index should
                          be added to the classification or not, defaults to
                          True.
    :type add_group_idx: bool, optional

    :param organic: Indicates whether soil is organic or not, defaults to False.
    :type organic: bool, optional

    :param clf_type: Used to indicate which type of soil classifier should be
                     used, defaults to None.
    :type clf_type: ClfType | str

    :raises ValueError: Raises ValueError if ``clf_type`` is  not supported or
                        None
    :raises ValueError: Raises ValueError if ``sand`` is not provided for
                        :class:`USCS` classification.
    """
    msg = ErrorMsg(param_name="clf_type",
                   param_value=clf_type,
                   symbol="in",
                   param_value_bound=list(ClfType))

    if clf_type is None:
        raise ValidationError(msg)

    try:
        clf_type = ClfType(str(clf_type).casefold())
    except ValueError as e:
        raise ValidationError(msg) from e

    atterberg_lmts = AtterbergLimits(liquid_limit=liquid_limit,
                                     plastic_limit=plastic_limit)

    if clf_type == ClfType.AASHTO:
        clf = AASHTO(atterberg_limits=atterberg_lmts,
                     fines=fines,
                     add_group_idx=add_group_idx)
        return clf

    # USCS classification
    if not sand:
        msg = ErrorMsg(param_name="sand",
                       param_value=sand,
                       msg="sand must be specified for USCS classification")
        raise ValidationError(msg)

    psd = PSD(fines=fines, sand=sand, d_10=d_10, d_30=d_30, d_60=d_60)
    clf = USCS(atterberg_limits=atterberg_lmts, psd=psd, organic=organic)

    return clf



from abc import ABC, abstractmethod

from geolysis.foundation import FoundationSize
from geolysis.utils import exceptions as exc
from geolysis.utils import validators


class AllowableBearingCapacity(ABC):
    #: Maximum tolerable foundation settlement (mm).
    MAX_TOL_SETTLEMENT = 25.4

    def __init__(self, corrected_spt_n_value: float,
                 tol_settlement: float,
                 foundation_size: FoundationSize) -> None:
        self.corrected_spt_n_value = corrected_spt_n_value
        self.tol_settlement = tol_settlement
        self.foundation_size = foundation_size

    @property
    def corrected_spt_n_value(self) -> float:
        """Statistical average of corrected SPT N-value."""
        return self._corrected_spt_n_value

    @corrected_spt_n_value.setter
    @validators.ge(0.0)
    def corrected_spt_n_value(self, val: float) -> None:
        self._corrected_spt_n_value = val

    @property
    def tol_settlement(self) -> float:
        """Tolerable settlement of foundation (mm)."""
        return self._tol_settlement

    @tol_settlement.setter
    @validators.le(25.4, exc_type=exc.SettlementError)
    def tol_settlement(self, tol_settlement: float) -> None:
        self._tol_settlement = tol_settlement

    def _sr(self) -> float:
        """Calculate the settlement ratio."""
        return self.tol_settlement / self.MAX_TOL_SETTLEMENT

    def _fd(self) -> float:
        """Calculate the depth factor."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        return min(1.0 + 0.33 * depth / width, 1.33)

    @abstractmethod
    def bearing_capacity(self): ...
